/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Guitone.h"
#include "WorkspaceWindow.h"
#include "DatabaseWindow.h"
#include "Preferences.h"
#include "Settings.h"

#ifdef Q_WS_MACX
#include "CocoaUtil.h"
#endif

#include <QMessageBox>
#include <QTimer>
#include <QMutexLocker>
#include <QFileInfo>
#include <QFileDialog>
#include <QDesktopWidget>
#include <QFileOpenEvent>

Guitone::Guitone(int & argc, char** argv)
  : QApplication(argc, argv), openPrompt(0),
    monotoneManager(0), somethingLoaded(false)
{
#ifndef Q_WS_MACX
    updateDialog = 0;
#else
    CocoaUtil::initialize();
#endif

    setQuitOnLastWindowClosed(false);

    // check for updates immediatly on launch
    if (Settings::getBool("CheckForUpdates", true))
    {
        checkForApplicationUpdates(true);
    }

    monotoneManager = new MonotoneManager();

    // Try to load something given on the command line
    // or found in one of the recent items lists as soon as
    // the manager got a valid binary path set
    // We need a queued, rather than direct connection to give
    // Qt the chance to process other events as well, so we
    // do receive f.e. FileOpen events properly
    connect(
        this, SIGNAL(monotoneBinaryPathSet()),
        this, SLOT(loadSomething()),
        Qt::QueuedConnection
    );

    QTimer::singleShot(0, this, SLOT(setMonotoneBinaryPath()));
}

Guitone::~Guitone()
{
    delete monotoneManager;
    I(openWindows.size() == 0);
    if (openPrompt) delete openPrompt;
#ifndef Q_WS_MACX
    if (updateDialog) delete updateDialog;
#endif
}

void Guitone::setMonotoneBinaryPath()
{
    if (monotoneManager->monotoneBinaryPathSet())
    {
        W("monotone binary path already set");
    }

    QString installedVersion;
    if (!monotoneManager->setMonotoneBinaryPath(
            Settings::getMtnBinaryPath(),
            Settings::getBool("RelaxedVersionCheck", false),
            installedVersion))
    {
        QMessageBox::critical(
            NULL,
            tr("Error"),
            tr("The path to the monotone binary is either invalid "
               "or points to a version of monotone with which guitone can't "
               "work with. Guitone requires monotone with an interface version "
               "between '%1' and '%2'. Your installation has the interface "
               "version '%3'.")
                   .arg(MonotoneManager::MinInterfaceVersion)
                   .arg(MonotoneManager::MaxInterfaceVersion)
                   .arg(installedVersion),
            QMessageBox::Ok, 0, 0
        );

        // note: the pref dialog tries to set the mtn binary path on its
        // own again and can only be accepted successfully, if the user
        // enters/selects a valid monotone binary path
        Preferences dialog(NULL);

        // the user has rejected the dialog, i.e. made no new settings
        if (dialog.exec() == QDialog::Rejected)
        {
            return;
        }
    }
    D("binary path was set");
    emit monotoneBinaryPathSet();
}

void Guitone::loadSomething()
{
    // if something was already loaded, skip this
    if (somethingLoaded) return;

    I(monotoneManager->monotoneBinaryPathSet());

    D("checking command line arguments and file open requests");

    QStringList args = arguments();
    for (int i=1, j=args.size(); i<j; i++)
    {
        // we're ok if any of these arguments could be loaded
        somethingLoaded = somethingLoaded || tryLoadSomething(args.at(i));
    }

    if (somethingLoaded) return;

    // if no command line arguments have been given or
    // they have been invalid or anything else went wrong,
    // try to load the recently used workspace/database
    D("trying to load recent workspace");
    QStringList workspaces = Settings::getItemList("RecentWorkspaceList");
    somethingLoaded = workspaces.size() > 0 && tryLoadSomething(workspaces.at(0));

    if (somethingLoaded) return;

    D("trying to load recent database");
    QStringList databases = Settings::getItemList("RecentDatabaseList");
    somethingLoaded = databases.size() > 0 && tryLoadSomething(databases.at(0));

    if (somethingLoaded) return;

    // if still nothing is loaded, prompt the user to load a workspace or database
    openPrompt = new OpenPrompt(NULL);

    connect(
        openPrompt, SIGNAL(loadDatabase(const QString &)),
        this, SLOT(loadDatabase(const QString &))
    );

    connect(
        openPrompt, SIGNAL(loadWorkspace(const QString &)),
        this, SLOT(loadWorkspace(const QString &))
    );

    connect(
        openPrompt, SIGNAL(quitApp()),
        this, SLOT(quit())
    );

    openPrompt->show();
}

bool Guitone::tryLoadSomething(const QString & path)
{
    bool loaded = true;
    try
    {
        loadFromPath(path);
    }
    catch (GuitoneException e)
    {
        W(QString("loading of '%1' failed: %2").arg(path).arg(e));
        loaded = false;
    }
    D(QString("loading of '%1' succeeded").arg(path));

    return loaded;
}

// this code is borrowed and adapted from QDesigner - thanks to the Trolls!
bool Guitone::event(QEvent * ev)
{
    bool eaten;
    switch (ev->type())
    {
        case QEvent::FileOpen:
        {
            D("got FileOpen event");
            QString file = static_cast<QFileOpenEvent *>(ev)->file();

            I(monotoneManager->monotoneBinaryPathSet());

            if (tryLoadSomething(file))
            {
                somethingLoaded = true;
            }
            eaten = true;
            break;
        }
        default:
            eaten = QApplication::event(ev);
            break;
    }
    return eaten;
}

void Guitone::loadFromPath(const QString & path)
{
    QMutexLocker locker(&lock);

    try
    {
        QFileInfo fileInfo(path);
        if (!fileInfo.exists())
            throw GuitoneException(tr("Non-existant path '%1'").arg(path));

        MainWindow * wnd;

        if (fileInfo.isDir())
            wnd = new WorkspaceWindow();
        else if (fileInfo.isFile())
            wnd = new DatabaseWindow();
        else
            throw GuitoneException(tr("Path is neither a file nor a directory"));

        wnd->init();

        try
        {
            wnd->load(path);
        }
        catch (GuitoneException e)
        {
            delete wnd;
            throw e;
        }

        openWindows.append(wnd);
        ensureCascadedWindowPlacement(wnd);
        wnd->show();

        triggerUpdateWindowList();
        emit updateRecentLists();
    }

    catch (GuitoneException e)
    {
        Settings::removeItemFromList("RecentWorkspaceList", path);
        Settings::removeItemFromList("RecentDatabaseList", path);
        emit updateRecentLists();
        throw e;
    }
}

void Guitone::loadWorkspace(const QString & path)
{
    try
    {
        loadFromPath(path);
    }
    catch (GuitoneException e)
    {
        QMessageBox::critical(
            NULL,
            tr("Failed to load workspace"),
            tr("The workspace could not be loaded:\n%1").arg(e),
            QMessageBox::Ok
        );
        return;
    }

    // we don't need this anymore
    if (openPrompt)
    {
        openPrompt->hide();
        delete openPrompt;
        openPrompt = 0;
    }
}

void Guitone::loadDatabase(const QString & path)
{
    try
    {
        loadFromPath(path);
    }
    catch (GuitoneException e)
    {
        QMessageBox::critical(
            NULL,
            tr("Failed to load database"),
            tr("The database could not be loaded:\n%1").arg(e),
            QMessageBox::Ok
        );
        return;
    }

    // we don't need this anymore
    if (openPrompt)
    {
        openPrompt->hide();
        delete openPrompt;
        openPrompt = 0;
    }
}

void Guitone::windowClosed(MainWindow * wnd)
{
    QMutexLocker locker(&lock);
    int i = openWindows.indexOf(wnd);
    I(i != -1);
    openWindows.removeAt(i);
    delete wnd;

    if (openWindows.size() == 0)
    {
        quit();
    }

    triggerUpdateWindowList();
}

void Guitone::triggerUpdateWindowList()
{
    QStringList list;
    foreach (MainWindow * win, openWindows)
    {
        list.append(win->windowTitle());
    }
    emit updateWindowList(list);
}

bool Guitone::closeAllWindows()
{
    // do we need to close any windows?
    if (openWindows.size() > 0)
    {
        // let all open windows close themselves to allow them cleaning up
        // stuff and saving settings
        // note that if the last window is closed, quit() is called again,
        // so we return here after closing all windows to avoid recursion
        // problems
        foreach (MainWindow * wnd, openWindows)
        {
            // QWidget::close() returns true if the widget could be closed,
            // i.e. hasn't ignored the close event
            if (!wnd->close()) return false;
        }
    }
    return true;
}

void Guitone::quit()
{
    if (!closeAllWindows()) return;
    Settings::sync();
    QApplication::quit();
}

void Guitone::checkForApplicationUpdates(bool silent)
{
#ifdef Q_WS_MACX
    CocoaUtil::checkForUpdates(silent);
#else
    if (!updateDialog)
    {
        updateDialog = new ApplicationUpdate(NULL);
    }

    if (!updateDialog->updateAvailable())
    {
        if (!silent)
        {
            QMessageBox::information(
                NULL,
                tr("No updates available"),
                tr("Your version of guitone (%1) is already up-to-date.")
                    .arg(GUITONE_VERSION),
                QMessageBox::Ok
            );
        }
        return;
    }
    updateDialog->show();
#endif
}

/*!
    Ensures somewhat that new windows do not overdraw current ones
    by adding a little x/y offset the original position of the window
    opened before this window
*/
void Guitone::ensureCascadedWindowPlacement(MainWindow * window)
{
    int curIdx = openWindows.indexOf(window);

    if (curIdx > 0)
    {
        MainWindow * prevWnd = openWindows.at(curIdx - 1);
        I(prevWnd);

        QDesktopWidget * desk = desktop();

        int cascade = 20;
        QRect geom = desk->availableGeometry(prevWnd);
        int newX = prevWnd->x() + cascade;
        int newY = prevWnd->y() + cascade;

        if (newX + window->width() > geom.right() ||
            newY + window->height() > geom.bottom())
        {
            newX = geom.x();
            newY = geom.y();
        }
        window->move(newX, newY);
    }
}

void Guitone::activateWindow(int idx)
{
    if (idx < 0 || idx > openWindows.size() - 1)
        return;

    MainWindow * wnd = openWindows.at(idx);
    wnd->activateWindow();
    wnd->raise();
}

void Guitone::raiseAllWindows()
{
    for (int i=0, j=openWindows.size(); i<j; i++)
    {
        MainWindow * wnd = openWindows.at(i);
        wnd->activateWindow();
        wnd->raise();
    }
}

