/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Branches.h"

Branches::Branches(QObject * parent, const DatabaseFile & db, bool display_tree)
  : QAbstractItemModel(parent), AutomateCommand(0),
    tree(display_tree), databaseFile(db)
{
    builder = 0;
}

Branches::~Branches()
{
    if (builder)
        delete builder;
}

void Branches::readBranches()
{
    branches.clear();
    reset();

    MonotoneTask task(QStringList() << "branches");
    AutomateCommand::enqueueDatabaseTask(databaseFile, task);
}

void Branches::processTaskResult(const MonotoneTask & task)
{
    if (task.getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task.getOutputUtf8()));
        return;
    }

    //
    // The ToolTipRole is used in two regards here:
    //
    //  a) if the branch model is displayed as tree of branches
    //     a user can hover over an individual branch part and see
    //     the complete branch name in the tooltip
    //  b) since we use a qstandarditemmodel internally to store the
    //     actual string data of each branch, we need to use some role
    //     anyways to store the full branch name internally to be able
    //     to use it for further processing, so we re-use the tooltip
    //     role there again to query these info
    //
    // Of course this will break badly as soon as we want to put more
    // info into the tooltip than just the actual branch name...
    //
    branches.setHorizontalHeaderItem(0, new QStandardItem(tr("Branches")));
    if (tree)
    {
        QStandardItem * root = new QStandardItem();
        root->setData("*", Qt::DisplayRole);
        root->setData("*", Qt::ToolTipRole);
        branches.appendRow(root);

        builder = new TreeBuilder(root, this);
        builder->addList(task.getOutputUtf8());
    }
    else
    {
        QStringList branchList = task.getOutputUtf8()
            .split("\n", QString::SkipEmptyParts);
        foreach (QString branch, branchList)
        {
            QStandardItem * item = new QStandardItem();
            item->setData(branch, Qt::DisplayRole);
            item->setData(branch, Qt::ToolTipRole);
            branches.appendRow(item);
        }
    }

    reset();
    emit branchesRead();
}

int Branches::columnCount(const QModelIndex & parent) const
{
    QStandardItem * item = branchFromIndex(parent);
    if (item)
    {
        return item->columnCount();
    }
    return branches.columnCount();
}

QVariant Branches::data(const QModelIndex & index, int role) const
{
    QStandardItem * item = branchFromIndex(index);
    if (item)
    {
        return item->data(role);
    }
    return QVariant();
}

QVariant Branches::headerData(int section, Qt::Orientation orientation, int role) const
{
    return branches.headerData(section, orientation, role);
}

int Branches::rowCount(const QModelIndex & parent) const
{
    QStandardItem * item = branchFromIndex(parent);
    if (item)
    {
        return item->rowCount();
    }
    return branches.rowCount();
}

QModelIndex Branches::index(int row, int column, const QModelIndex & parent) const
{
    return createIndex(row, column, getBranchItem(row, column, parent));
}

QModelIndex Branches::parent(const QModelIndex & index) const
{
    QStandardItem * item = branchFromIndex(index);
    if (item)
    {
        QStandardItem * parent = item->parent();
        if (parent)
        {
            return createIndex(parent->row(), parent->column(), parent);
        }
    }
    return QModelIndex();
}

QStandardItem * Branches::getBranchItem(int row, int column, const QModelIndex & parent) const
{
    if (parent.isValid())
    {
        return static_cast<QStandardItem *>(parent.internalPointer())->child(row, column);
    }
    return branches.item(row, column);
}

QStandardItem * Branches::branchFromIndex(const QModelIndex & index) const
{
    if (index.isValid())
    {
        return static_cast<QStandardItem *>(index.internalPointer());
    }
    return 0;
}

