/***************************************************************************
 *   Copyright (C) 2006 by Jean-Louis Fuchs                                 *
 *   ganwell@fangorn.ch                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "ChangesetModel.h"
#include <QFont>

ChangesetModel::ChangesetModel(QObject * parent, const DatabaseFile & db)
  : QAbstractItemModel(parent), databaseFile(db), currentBranch(""), branchLog(0)
{}

ChangesetModel::~ChangesetModel() {}

void ChangesetModel::receiveRevisions(bool all)
{
    int i = branchMap[currentBranch].certsRead;
    int to = branchMap[currentBranch].revisions.count();
    int c = 0;

    while (i < to && (c < revPerReceive || all))
    {
        QString revision = branchMap[currentBranch].revisions[i];

        if (!changesetMap.contains(revision))
        {
            ChangesetCerts * certs = new ChangesetCerts(this, databaseFile);
            connect(
                certs, SIGNAL(changesetReady(const Changeset &)),
                this, SLOT(changesetReady(const Changeset &))
            );
            certs->readCerts(revision);
            c++;
        }
        i++;
    }
}

void ChangesetModel::setBranch(const QString & branch)
{
    currentBranch = branch;

    if (branchMap[currentBranch].revisions.count() == 0)
    {
        branchMap[currentBranch].certsRead = 0;

        if (branchLog)
        {
            disconnect(
                branchLog, SIGNAL(branchLogRead(const QString &, const QStringList &)),
                this, SLOT(branchLogRead(const QString &, const QStringList &))
            );
            delete branchLog;
        }

        branchLog = new GetBranchLog(this, databaseFile, branch);
        connect(
            branchLog, SIGNAL(branchLogRead(const QString &, const QStringList &)),
            this, SLOT(branchLogRead(const QString &, const QStringList &))
        );
    }
    else
    {
        // just reset the view if we already have queried data
        // for this particular branch
        reset();
    }
}

void ChangesetModel::branchLogRead(const QString & branch, const QStringList & revisions)
{
    branchMap[branch].revisions += revisions;

    if (branch == currentBranch)
    {
        receiveRevisions(false);
    }
}

void ChangesetModel::changesetReady(const Changeset & change)
{
    beginInsertRows(
        QModelIndex(),
        branchMap[currentBranch].certsRead,
        branchMap[currentBranch].certsRead
    );
    changesetMap[change.revision] = change;
    branchMap[currentBranch].certsRead += 1;
    endInsertRows();
}

int ChangesetModel::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 4;
}

QVariant ChangesetModel::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    if (role == Qt::FontRole && index.column() == 3)
    {
        QFont font;
        font.setStyleHint(QFont::Courier);
        font.setFamily("Courier");
        return QVariant(font);
    }

    if (role == Qt::UserRole)
    {
        BranchInfo branchInfo = branchMap[currentBranch];
        int row = index.row();
        QString rev = branchInfo.revisions[row];
        if (!changesetMap.contains(rev)) return QVariant();
        return QVariant(changesetMap[rev].htmllog());
    }

    if (role == Qt::DisplayRole)
    {
        BranchInfo branchInfo = branchMap[currentBranch];
        int row = index.row();

        if (branchInfo.revisions.count() > row)
        {
            int col = index.column();
            QString rev = branchInfo.revisions[row];

            if (!changesetMap.contains(rev)) return QVariant();

            switch(col)
            {
                case(0):
                    return QVariant(changesetMap[rev].date);
                case(1):
                    return QVariant(changesetMap[rev].author);
                case(2):
                    return QVariant(changesetMap[rev].plainlog());
                case(3):
                    return QVariant(rev);
            }
        }
    }

    return QVariant();
}

Qt::ItemFlags ChangesetModel::flags(const QModelIndex &index) const
{
    if (index.isValid())
    {
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
    }
    return 0;
}

QVariant ChangesetModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        switch(section)
        {
        case(0):
            return QVariant(tr("Date"));
        case(1):
            return QVariant(tr("Author"));
        case(2):
            return QVariant(tr("Changelog"));
        case(3):
            return QVariant(tr("Revision ID"));
        }
    }
    return QVariant();
}

int ChangesetModel::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    if (branchMap[currentBranch].revisions.count() > 0)
    {
        return branchMap[currentBranch].certsRead;
    }
    return 0;
}

QModelIndex ChangesetModel::index(int row, int column, const QModelIndex & parent) const
{
    if (!hasIndex(row, column, parent))
    {
        return QModelIndex();
    }

    return createIndex(row, column, 0);
}

QModelIndex ChangesetModel::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

