/***************************************************************************
 *   Copyright (C) 2006 by Jean-Louis Fuchs                                 *
 *   ganwell@fangorn.ch                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; withhtmllog even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef CHANGESETMODEL_H
#define CHANGESETMODEL_H

#include "Certs.h"
#include "GetBranchLog.h"

#include <QDateTime>
#include <QRegExp>

struct BranchInfo
{
    QStringList revisions;
    int certsRead;
};

struct Changeset
{
    QString revision;
    QDateTime date;
    QString author;
    QString changelog;

    QString htmllog() const
    {
        QStringList lines = changelog.split(QRegExp("(\\r\\n|\\n|\\n\\r)"));
        QString htmllog;
        bool inList = false;

        foreach (QString line, lines)
        {
            if (line.isEmpty())
            {
                if (inList)
                {
                    htmllog += "</li>\n</ul>\n";
                    inList = false;
                    continue;
                }

                htmllog += "<br/>\n";
                continue;
            }

            bool listStart = (0 == line.indexOf(QRegExp("^[*-]")));
            if (listStart)
            {
                if (inList)
                {
                    htmllog += "</li>\n";
                }
                else
                {
                    inList = true;
                    htmllog += "<ul>\n";
                }

                htmllog += "<li>" + line.mid(1).trimmed();
                continue;
            }

            if (inList)
            {
                htmllog += " " + line.trimmed();
                continue;
            }

            htmllog += line.trimmed() + "<br/>\n";
        }

        if (inList)
        {
            htmllog += "</li>\n</ul>\n";
        }

        htmllog = htmllog
            .replace(QRegExp("(\\b)([_])(\\w+)\\2"), "\\1<u>\\2\\3\\2</u>")
            .replace(QRegExp("(\\b)([*])(\\w+)\\2"), "\\1<b>\\2\\3\\2</b>")
            .replace(QRegExp("(https?://[^() ]+)"),   "<a href=\"\\1\">\\1</a>")
            .replace(QRegExp("\\t"), "&nbsp;&nbsp;&nbsp;&nbsp;");

        return htmllog;
    }

    QString plainlog() const
    {
        QString plainlog = changelog;
        return plainlog.replace(QRegExp("\\s+"), " ");
    }
};

class ChangesetCerts : public Certs
{
    Q_OBJECT
public:
    ChangesetCerts(QObject * parent, const DatabaseFile & db) : Certs(parent, db)
    {
        connect(
            this, SIGNAL(certsRead()),
            this, SLOT(packIntoChangeset())
         );
    }
    void readCerts(const QString & rev)
    {
        revision = rev;
        Certs::readCerts(rev);
    }

signals:
    void changesetReady(const Changeset &);

private slots:
    void packIntoChangeset()
    {
        Changeset change;
        change.revision = revision;

        // FIXME: there is no support for multiple certs with the same
        // key, i.e. accidential clean merges or multiple author certs
        for (int i=0, j=certs->size(); i<j; i++)
        {
            if (certs->at(i).name == "author")
            {
                change.author = certs->at(i).value;
                continue;
            }
            if (certs->at(i).name == "changelog")
            {
                change.changelog = certs->at(i).value;
                continue;
            }
            if (certs->at(i).name == "date")
            {
                change.date = QDateTime::fromString(certs->at(i).value, "yyyy-MM-ddThh:mm:ss");
                continue;
            }
        }

        emit changesetReady(change);
    }

private:
    QString revision;
};

typedef QMap<QString, Changeset> ChangesetMap;
typedef QMap<QString, BranchInfo> BranchMap;

class ChangesetModel : public QAbstractItemModel
{
    Q_OBJECT

public:
    ChangesetMap branchModel;
    ChangesetModel(QObject *, const DatabaseFile &);
    ~ChangesetModel();

    void setBranch(const QString & branch);
    void receiveRevisions(bool);

    // needed Qt Model methods
    QVariant data(const QModelIndex &, int) const;
    Qt::ItemFlags flags(const QModelIndex &) const;
    QVariant headerData(int, Qt::Orientation, int) const;
    QModelIndex index(int, int, const QModelIndex &) const;
    QModelIndex parent(const QModelIndex &) const;
    int rowCount(const QModelIndex &) const;
    int columnCount(const QModelIndex &) const;

private:
    static const int revPerReceive = 50;

    DatabaseFile databaseFile;
    QString currentBranch;
    BranchMap branchMap;
    ChangesetMap changesetMap;
    GetBranchLog * branchLog;

private slots:
    void branchLogRead(const QString &, const QStringList &);
    void changesetReady(const Changeset &);
};

#endif

