/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "GetContentChanged.h"
#include "BasicIOParser.h"
#include "MonotoneUtil.h"

#include <QFont>

GetContentChanged::GetContentChanged(QObject * parent, const DatabaseFile & db)
    : QAbstractItemModel(parent), AutomateCommand(0), databaseFile(db)
{}

GetContentChanged::~GetContentChanged()
{
    revisions.clear();
    pathInRevision.clear();
}

//
// The basic flow is the following:
//
//    [ workspace parent revision ]  [ path in this revision ]
//                               \     /
//                                \   /
//                                 \ /
//                                  |
//                                  V
//     |---------------> [  get content changed ]
//     |                            |
//     |                            V
//     |            [ add the marked rev(s) to the list ]
//     |                            |
//     |                            V
//     |       [ get corresponding path for each marked rev ]
//     |                            |
//     |                            V
//     |---- < ---- [ get parents of each marked rev ]  -> stop if no parents
//     |                            |
//     |                            V
//     |--- < --- [ get corresponding path for each parent ] -> stop if no path
//
void GetContentChanged::readChanges(const QString & path, const QString & rev)
{
    revisions.clear();
    pathInRevision.clear();

    // reset the view
    reset();

    // FIXME: we assume that the given path is part of this revision!
    startRev = rev;
    startPath = path;

    queryContentChanged(startRev, startPath);
}

void GetContentChanged::queryContentChanged(const QString & rev, const QString & path)
{
    MonotoneTask task(QStringList() << "get_content_changed" << rev << path);
    AutomateCommand::enqueueDatabaseTask(databaseFile, task);
}

void GetContentChanged::queryParents(const QString & rev)
{
    MonotoneTask task(QStringList() << "parents" << rev);
    AutomateCommand::enqueueDatabaseTask(databaseFile, task);
}

void GetContentChanged::queryCorrespondingPath(const QString & rev, const QString & path, const QString & par)
{
    MonotoneTask task(QStringList() << "get_corresponding_path" << rev << path << par);
    AutomateCommand::enqueueDatabaseTask(databaseFile, task);
}

void GetContentChanged::processTaskResult(const MonotoneTask & task)
{
    if (task.getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task.getOutputUtf8()));
        return;
    }

    QString current = task.getArguments().at(0);
    QString output = task.getOutputUtf8();

    if (current == "parents")
    {

        if (output.isEmpty())
        {
            reset();
            emit rootReached();
            return;
        }

        QStringList parents = output.split(
             '\n', QString::SkipEmptyParts
        );

        foreach (QString par, parents)
        {
            // if we checked this revision already, skip it
            if (pathInRevision.contains(par))
            {
                continue;
            }

            revsForPathStack.enqueue(par);
            queryCorrespondingPath(startRev, startPath, par);
        }
        return;
    }

    if (current == "get_content_changed")
    {
        // since we're looking for a file in a certain revision
        // _beforehand_ via get_corresponding_path, we should always
        // get a stanza out here
        BasicIOParser parser(output);
        I(parser.parse());
        StanzaList stanzas = parser.getStanzas();
        foreach (Stanza st, stanzas)
        {
            I(st.size() == 1);
            QString rev = st.at(0).hash;
            I(!rev.isNull());

            // if we have this particular revision already recorded,
            // skip it
            if (revisions.contains(rev)) continue;

            // append the revision to the ordered list
            revisions.append(rev);

            // check if we already know the path in that revision
            if (!pathInRevision.contains(rev))
            {
                // add the revision to the stack of revs which need
                // to get a valid path queried
                revsForPathStack.enqueue(rev);

                // query for the corresponding path
                queryCorrespondingPath(startRev, startPath, rev);
            }

            // query for the parents of this revision for the
            // next round
            queryParents(rev);
        }
        return;
    }

    if (current == "get_corresponding_path")
    {
        I(revsForPathStack.size() > 0);
        QString rev = revsForPathStack.dequeue();

        // if there is no corresponding path for the file in this
        // revision, we've reached the end of this development line
        if (output.isEmpty())
        {
            reset();
            emit endOfLineReached();
            return;
        }

        BasicIOParser parser(output);
        I(parser.parse());
        StanzaList stanzas = parser.getStanzas();
        I(stanzas.size() == 1);
        Stanza st = stanzas.at(0);
        I(st.size() == 1);
        StanzaEntry en = st.at(0);
        I(en.sym == "file" && en.vals.size() == 1);
        QString path = en.vals.at(0);

        // we're also inserting a lot of uninteresting (revision, path)
        // tuples here, i.e. for every parent revision of a marked node,
        // however the only list that counts is revisions
        pathInRevision.insert(rev, path);

        // try to get the next marked node
        queryContentChanged(rev, path);
        return;
    }

    I(false);
}

int GetContentChanged::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant GetContentChanged::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    int col = index.column();

    if (role == Qt::FontRole && col == 0)
    {
        QFont font;
        font.setStyleHint(QFont::Courier);
        font.setFamily("Courier");
        return QVariant(font);
    }

    if (role == Qt::DisplayRole)
    {
        int row = index.row();
        if (row >= revisions.size()) return QVariant();
        QString rev = revisions.at(row);
        I(pathInRevision.contains(rev));

        switch (col)
        {
            case 0: return QVariant(rev);
            case 1: return QVariant(pathInRevision.value(rev));
        }
        return QVariant();
    }

    return QVariant();
}

Qt::ItemFlags GetContentChanged::flags(const QModelIndex & index) const
{
    if (index.isValid())
    {
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
    }
    return 0;
}

QVariant GetContentChanged::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(tr("Revision ID"));
            case 1: return QVariant(tr("Path in this revision"));
        }
    }
    return QVariant();
}

int GetContentChanged::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return revisions.size();
}

QModelIndex GetContentChanged::index(int row, int column, const QModelIndex & parent) const
{
    if (!hasIndex(row, column, parent))
    {
        return QModelIndex();
    }

    return createIndex(row, column, 0);
}

QModelIndex GetContentChanged::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

