/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "GetFile.h"
#include "MonotoneUtil.h"

#include <QBrush>
#include <QFont>

GetFile::GetFile(QObject * parent, const DatabaseFile & db)
  : QAbstractItemModel(parent), AutomateCommand(0), databaseFile(db)
{}

GetFile::~GetFile() {}

void GetFile::readFileByName(const QString & fileName, const QString & rev)
{
    QStringList cmd;
    cmd << "get_file_of" << fileName;

    QStringList opts;
    opts << "r" << rev;

    readFile(cmd, opts);
}

void GetFile::readFileById(const QString & fileID)
{
    QStringList cmd;
    cmd << "get_file" << fileID;

    readFile(cmd, QStringList());
}

void GetFile::readFile(QStringList cmd, QStringList opts)
{
    // reset attached views
    reset();
    MonotoneTask task(cmd, opts);
    AutomateCommand::enqueueDatabaseTask(databaseFile, task);
}

void GetFile::processTaskResult(const MonotoneTask & task)
{
    if (task.getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task.getOutputUtf8()));
        return;
    }

    // FIXME: if the file contents is misinterpreted as utf-8, we have a
    // problem here. The only thing we could do here is to let the user
    // select the charset of the file and re-read it afterwards
    QString output = task.getOutputUtf8();

    QRegExp rx("\\0x00");
    if (rx.indexIn(output) != -1)
    {
        // FIXME: signal this anywhere...
        W("File is binary");
        return;
    }

    QStringList lines(output.split(QRegExp("\\n|\\r\\n")));
    // skip the last line since this contains a newline from automate
    for (int i=0, s=lines.size() - 1; i<s; i++)
    {
        fileContents.push_back(
            new ContentLine(lines.at(i),ContentLine::Unchanged)
        );
    }

    reset();

    emit fileRead();
}

void GetFile::applyDiff(Diff * diff)
{
    int lineOffset = 0;

    for (int i=0, s=diff->hunks.size(); i<s; i++)
    {
        DiffHunk * hunk = diff->hunks.at(i);
        int lineStart = hunk->leftStart + lineOffset;

        for (int j=0, t=hunk->lines.size(); j<t; j++)
        {
            DiffLine * line = hunk->lines.at(j);

            // skip unchanged lines
            if (line->state == DiffLine::Unchanged)
            {
                lineStart++;
                continue;
            }

            if (line->state == DiffLine::Removed)
            {
                fileContents.value(lineStart-1)->marker = ContentLine::Removed;
                lineStart++;
                continue;
            }

            I(line->state == DiffLine::Added);

            fileContents.insert(
                lineStart - 1,
                new ContentLine(line->content, ContentLine::Added)
            );
            lineOffset++;
            lineStart++;
        }
    }

    reset();
}

QModelIndex GetFile::getNextGroup(const QModelIndex & base, bool recurse)
{
    int startRow = 0;
    ContentLine::Marker lastMarker = ContentLine::Unchanged;
    if (base.isValid() && base.row() < fileContents.size())
    {
        startRow = base.row();
        lastMarker = fileContents.at(startRow)->marker;
    }

    for (int i=startRow, j=fileContents.size(); i<j; i++)
    {
        ContentLine * line = fileContents.at(i);
        if (line->marker != lastMarker && line->marker != ContentLine::Unchanged)
        {
            return index(i, 0, QModelIndex());
        }
        lastMarker = line->marker;
    }

    // if we already recursed, stop here
    if (recurse) return QModelIndex();
    // try to get the first group
    return getNextGroup(QModelIndex(), true);
}

QModelIndex GetFile::getPrevGroup(const QModelIndex & base, bool recurse)
{
    int startRow = fileContents.size() - 1;
    ContentLine::Marker lastMarker = ContentLine::Unchanged;
    if (base.isValid() && base.row() < fileContents.size())
    {
        startRow = base.row();
        lastMarker = fileContents.at(startRow)->marker;
    }

    bool foundLast = false;
    for (int i=startRow; i>=0; i--)
    {
        ContentLine * line = fileContents.at(i);

        // did we already found the last line of a group of similar lines?
        if (!foundLast)
        {
            if (line->marker != lastMarker && line->marker != ContentLine::Unchanged)
            {
                foundLast = true;
            }
            lastMarker = line->marker;
            continue;
        }

        if (foundLast)
        {
            // if the marker just changed, the previous line was our
            // first line
            if (line->marker != lastMarker)
            {
                return index(i + 1, 0, QModelIndex());
            }
        }
    }

    // if we've already recursed, stop here
    if (recurse) return QModelIndex();
    // try to get the last group
    return getPrevGroup(QModelIndex(), true);
}

int GetFile::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant GetFile::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    int row = index.row();
    if (row >= fileContents.size()) return QVariant();

    ContentLine * line = fileContents.at(row);

    if (role == Qt::DisplayRole)
    {
        switch (index.column())
        {
            case 0: return QVariant(row + 1);
            case 1: return QVariant(line->content);
            default: I(false);
        }
    }
    else
    if (role == Qt::BackgroundRole)
    {
        switch (line->marker)
        {
            case ContentLine::Added: return QVariant(QBrush(Qt::green));
            case ContentLine::Removed: return QVariant(QBrush(Qt::red));
            default: return QVariant();
        }
    }
    else
    if (role == Qt::FontRole && index.column() == 1)
    {
        QFont font;
        font.setStyleHint(QFont::Courier);
        font.setFamily("Courier");
        return QVariant(font);
    }
    else
    if (role == Qt::TextAlignmentRole && index.column() == 0)
    {
        return QVariant(Qt::AlignRight);
    }
    return QVariant();
}

Qt::ItemFlags GetFile::flags(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return Qt::ItemIsEnabled;
}

QVariant GetFile::headerData(int section, Qt::Orientation orientation, int role) const
{
    Q_UNUSED(orientation);

    if (role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(tr("Line"));
            case 1: return QVariant(tr("Content"));
        }
    }

    return QVariant();
}

int GetFile::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return fileContents.size();
}

QModelIndex GetFile::index(int row, int column, const QModelIndex& parent) const
{
    if (!hasIndex(row, column, parent))
    {
        return QModelIndex();
    }

    ContentLine * line = fileContents.at(row);
    return createIndex(row, column, line);
}

QModelIndex GetFile::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

