/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "GetRevision.h"
#include "vocab.h"

#include <QFont>
#include <QBrush>
#include <QTextStream>

GetRevision::GetRevision(QObject * parent)
    : QAbstractItemModel(parent), AutomateCommand(0)
{}

GetRevision::~GetRevision() {}

void GetRevision::readDatabaseRevision(const DatabaseFile & db, const QString & rev)
{
    revision.clear();
    MonotoneTask task( QStringList() << "get_revision" << rev);
    AutomateCommand::enqueueDatabaseTask(db, task);
}

void GetRevision::readWorkspaceRevision(const WorkspacePath & ws)
{
    revision.clear();
    MonotoneTask task( QStringList() << "get_revision");
    AutomateCommand::enqueueWorkspaceTask(ws, task);
}

void GetRevision::processTaskResult(const MonotoneTask & task)
{
    if (task.getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task.getOutputUtf8()));
        return;
    }

    BasicIOParser parser(task.getOutputUtf8());
    I(parser.parse());

    StanzaList list = parser.getStanzas();
    QString currentRevision;

    for (int i=0, size = list.size(); i < size; ++i)
    {
        Stanza stanza = list.at(i);
        Change change;
        bool found_change = false;

        for (int j=0, size2 = stanza.size(); j < size2; j++)
        {
            StanzaEntry entry = stanza.at(j);

            if (i == 0 && j == 0)
            {
                if (entry.sym != "format_version")
                {
                    W("First entry in first stanza should be used format version.");
                } else
                if (entry.vals.at(0) != "1")
                {
                    W("Format version is unequal to '1' - newer interface version?");
                }
                break;
            }

            if (j == 0 && entry.sym == "new_manifest")
            {
                I(!entry.hash.isNull());
                revision.new_manifest = entry.hash;
                break;
            }

            if (entry.sym == "old_revision")
            {
                I(!entry.hash.isNull());
                currentRevision = entry.hash;
                QList<Change> changelist;
                revision.changesAgainstParent.insert(currentRevision, changelist);
                break;
            }

            found_change = true;

            if (entry.sym == "delete")
            {
                I(entry.vals.size() == 1);
                change.type = Delete;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "rename")
            {
                I(entry.vals.size() == 1);
                change.type = Rename;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "add_dir")
            {
                I(entry.vals.size() == 1);
                change.type = AddDir;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "add_file")
            {
                I(entry.vals.size() == 1);
                change.type = AddFile;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "patch")
            {
                I(entry.vals.size() == 1);
                change.type = Patch;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "clear")
            {
                I(entry.vals.size() == 1);
                change.type = AttrClear;
                change.stanza = stanza;
                break;
            }

            if (entry.sym == "set")
            {
                I(entry.vals.size() == 1);
                change.type = AttrSet;
                change.stanza = stanza;
                break;
            }

            found_change = false;

            W(QString("Unknown symbol %1.").arg(entry.sym));
        }

        // check if we really processed an item entry
        if (!found_change) continue;
        I(revision.changesAgainstParent.contains(currentRevision));
        revision.changesAgainstParent[currentRevision].append(change);
    }

    // set the first as the default parent against which changes are displayed
    changesAgainstParent = revision.changesAgainstParent.keys().at(0);

    // reset the view
    reset();

    // signal that we've finished (whoever listens to that)
    emit revisionRead();
}

int GetRevision::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 2;
}

QVariant GetRevision::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
        return QVariant();

    if (!revision.changesAgainstParent.contains(changesAgainstParent))
        return QVariant();

    Change change =
        revision.changesAgainstParent[changesAgainstParent].at(index.row());

    if (role == Qt::DisplayRole)
    {
        switch (index.column())
        {
            case 0: return QVariant(change.getTypeString());
            case 1: return QVariant(change.getDisplayData());
            default: I(false);
        }
    }
    else
    if (role == Qt::BackgroundRole)
    {
        if (index.column() == 0)
        {
            return QVariant(QBrush(change.getTypeColor()));
        }
        return QVariant();
    }
    else
    if (role == Qt::FontRole && index.column() == 0)
    {
        QFont font;
        font.setPointSize(9);
        return QVariant(font);
    }
    else
    if (role == Qt::TextAlignmentRole && index.column() == 0)
    {
        return QVariant(Qt::AlignCenter);
    }
    return QVariant();
}

Qt::ItemFlags GetRevision::flags(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant GetRevision::headerData(int section, Qt::Orientation orientation, int role) const
{
    Q_UNUSED(orientation);
    if (role == Qt::DisplayRole)
    {
        switch (section)
        {
            case 0: return QVariant(tr("Type"));
            case 1: return QVariant(tr("Change"));
        }
    }

    return QVariant();
}

int GetRevision::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    if (!revision.changesAgainstParent.contains(changesAgainstParent))
        return 0;

    return revision.changesAgainstParent[changesAgainstParent].size();
}

QModelIndex GetRevision::index(int row, int column, const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    if (!revision.changesAgainstParent.contains(changesAgainstParent))
        return QModelIndex();

    if (row >= revision.changesAgainstParent[changesAgainstParent].size())
        return QModelIndex();

    return createIndex(row, column);
}

QModelIndex GetRevision::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

QStringList GetRevision::getParentRevisions() const
{
    return revision.changesAgainstParent.keys();
}

void GetRevision::showChangesAgainstParent(const QString & parent)
{
    QStringList parents = getParentRevisions();

    if (!parents.contains(parent)) return;

    changesAgainstParent = parent;
    reset();
}

