/***************************************************************************
*   Copyright (C) 2006 by Thomas Keller                                   *
*   me@thomaskeller.biz                                                   *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "Inventory.h"
#include "MonotoneUtil.h"
#include "BasicIOParser.h"
#include "Settings.h"

#include <QApplication>
#include <QSet>

const int Inventory::QueryLevel = 1;

Inventory::Inventory(QObject * parent)
  : QAbstractItemModel(parent), AutomateCommand(0), workspacePath()
{
    rootItem = new ModelItem("__root__");
    rootItem->setParent(rootItem);

    connect(
        this, SIGNAL(modelReset()),
        this, SLOT(setWorkspacePath())
    );
}

Inventory::~Inventory()
{
    delete rootItem;
}

void Inventory::setWorkspacePath(const WorkspacePath & ws)
{
    workspacePath = ws;
}

void Inventory::refresh()
{
    readInventory(QString());
}

bool Inventory::canFetchMore(const QModelIndex & parent) const
{
    // special handling for the root item
    if (!parent.isValid())
        return true;

    ModelItem * item = static_cast<ModelItem *>
        (parent.internalPointer());
    I(item);

    InventoryItem * invitem = dynamic_cast<InventoryItem *>
        (item);

    // not an inventory item
    if (!invitem)
    {
        D(QString("skipping %1 (not an inventory item)").arg(item->getLabel()));
        return false;
    }

    if (invitem->isExpanded(QueryLevel))
    {
        D(QString("skipping %1 (already expanded)").arg(item->getLabel()));
        return false;
    }

    return true;
}

void Inventory::fetchMore(const QModelIndex & parent)
{
    if (!parent.isValid())
        return;

    ModelItem * item = static_cast<ModelItem *>
        (parent.internalPointer());
    I(item);

    InventoryItem * invitem = dynamic_cast<InventoryItem *>
        (item);

    // not an inventory item
    if (!invitem)
    {
        D(QString("skipping %1 (not an inventory item)").arg(item->getLabel()));
        return;
    }

    // mtn currently allows to restrict on ignored, but not on unknown
    // paths, check that!
    if (invitem->hasStatus(InventoryItem::Unknown))
    {
        D(QString("skipping %1 (unknown path)").arg(invitem->getPath()));
        return;
    }

    // 'inventory' can only be restricted with existing file paths
    // and not with invalid file/directory combinations or missing paths
    // restricting to single files makes no sense for us here either
    if (invitem->getFSType() == InventoryItem::None)
    {
        D(QString("skipping %1 (not an existing path)").arg(invitem->getPath()));
        return;
    }

    // again, if this item is already expanded, don't expand it again
    if (invitem->isExpanded(QueryLevel)) return;

    readInventory(invitem->getPath());
}

void Inventory::readInventory(const QString & path)
{
    I(!workspacePath.isEmpty());

    QStringList cmd = QStringList() << "inventory";
    if (!path.isEmpty())
    {
        cmd << path;
    }

    QStringList opts;

    // if requested, just read one directory level ahead
    // (useful for big directory trees)
    if (Settings::getBool("ReadWorkspaceIncrementally", false))
    {
        opts << "depth" << QString::number(QueryLevel);
    }

    MonotoneTask task(cmd, opts);
    AutomateCommand::enqueueWorkspaceTask(workspacePath, task);
}

/*!
    @fixme
    The whole item adding procedurecould eventually make problems
    if inventory is executed with a path which was renamed. The
    inventory output then contains another stanza of a path which
    was not yet read in, so we have no plan where to tack this on.
    A possible resolution would be to enqueue a new inventory
    request with the baseDir right here, delete the node and
    continue with the next one
*/
void Inventory::processTaskResult(const MonotoneTask & task)
{
    if (task.getReturnCode() != 0)
    {
        if (task.getReturnCode() == 2)
        {
            emit invalidWorkspaceFormat(
                MonotoneUtil::stripMtnPrefix(task.getOutputUtf8())
            );
            return;
        }

        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task.getOutputUtf8()));
        return;
    }

    QString queriedPath = "";
    if (task.getArguments().size() > 1)
    {
        queriedPath = task.getArguments().at(1);
    }

    int queriedDepth = -1;
    if (task.getOptions().size() > 1)
    {
        queriedDepth = task.getOptions().at(1).toInt();
        if (!queriedPath.isEmpty())
        {
            queriedDepth += queriedPath.split("/", QString::SkipEmptyParts).size();
        }
    }

    BasicIOParser parser(task.getOutputUtf8());
    I(parser.parse());
    StanzaList stanzas = parser.getStanzas();

    ModelItem * parentItem = rootItem;
    if (itemMap.contains(queriedPath))
    {
        parentItem = itemMap.value(queriedPath)->parent();
    }

    //
    // For the lazy population code to work properly, Qt needs to get
    // all items added within their very own beginInsertRows and
    // endInsertRows. To prevent that we need to call those two for each
    // single item, we do not yet set the parent / child dependencies just
    // now, but record them temporarily in parentChildRelations.
    // If we've processed all stanzas, we rework these relations into
    // actual relations and notify Qt about that in insertRowsRecursive
    //
    QMap<ModelItem *, QList<ModelItem *> > parentChildRelations;
    QMap<QString, ModelItem *> tempItemMap;
    foreach (Stanza st, stanzas)
    {
        InventoryItem * item = new InventoryItem(st);

        QString path = item->getPath();
        int itemDepth = 0;
        if (!path.isEmpty())
        {
            itemDepth = path.split("/", QString::SkipEmptyParts).size();
        }

        // add a cdup item for this item if we either queried everything
        // already or the current directory level corresponds to the queried one
        if (item->isDirectory() && (queriedDepth == -1 || queriedDepth >= itemDepth))
        {
            PseudoItem * cdUp = new PseudoItem("..", PseudoItem::CdUp);
            parentChildRelations[item].append(cdUp);
            item->setExpanded();
        }

        QString baseDir = item->getBaseDirectory();

        ModelItem * parent = 0;
        if (tempItemMap.contains(baseDir))
            parent = tempItemMap.value(baseDir);
        else
            parent = parentItem;

        I(parent);
        parentChildRelations[parent].append(item);
        tempItemMap.insert(path, item);
    }

    insertRowsRecursive(parentItem, parentChildRelations);

    //
    // This is a bit of a hack: Since we replace the original item each
    // time we query new data, the view looses the viewport and indexes
    // for it. So what we need to do here is to notify the view that it
    // should set the viewport to the correct item again _after_ we've
    // read it (now) and we need to do this through an already existing
    // signal since this model is not directly connected to the view, but
    // through a proxy model, so we need to ensure that the proxy model
    // transports a known signal to the view.
    // The dataChanged signal is normally used for in-place editable
    // actions, so I really hope it has
    //  a) no side effects
    //  b) I never need it for its original purpose
    //
    if (queriedPath.isEmpty())
    {
        // we need to call reset() before we emit dataChanged,
        // otherwise the view doesn't know anything of the new
        // index we're going to present her afterwards
        reset();
    }

    I(itemMap.contains(queriedPath));
    QModelIndex index = indexFromItem(itemMap.value(queriedPath), 0);
    I(index.isValid());
    emit dataChanged(index, index);
}

QModelIndex Inventory::indexFromItem(ModelItem * item, int col) const
{
    if (item->isRoot())
    {
        return QModelIndex();
    }
    return createIndex(item->row(), col, reinterpret_cast<void *>(item));
}

void Inventory::setWorkspacePath()
{
    QList<ModelItem *> children = rootItem->getChildren();

    int workspaceRootRow = 0;

    I(children.size() > workspaceRootRow);

    QString displayPath = workspacePath;
    QString homePath = QDir::homePath();
    if (workspacePath.indexOf(homePath) == 0)
    {
        displayPath = QDir(homePath).relativeFilePath(workspacePath);
        displayPath.prepend("~/");
    }
    children[workspaceRootRow]->setLabel(displayPath);

    emit dataChanged(
        index(workspaceRootRow, 0, QModelIndex()),
        index(workspaceRootRow, 2, QModelIndex())
    );
}

QModelIndex Inventory::index(int row, int column, const QModelIndex & parent) const
{
    ModelItem * parentItem;

    if (!parent.isValid())
    {
        parentItem = rootItem;
    }
    else
    {
        parentItem = static_cast<ModelItem *>(parent.internalPointer());
        I(parentItem);
    }

    ModelItem * childItem = parentItem->child(row);

    if (childItem)
    {
        return createIndex(row, column, childItem);
    }

    return QModelIndex();
}

int Inventory::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 3;
}

QVariant Inventory::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
    I(item);
    return item->data(index.column(), role);
}

Qt::ItemFlags Inventory::flags(const QModelIndex & index) const
{
    if (!index.isValid()) return 0;

    QFlags<Qt::ItemFlag> flags = Qt::ItemIsSelectable | Qt::ItemIsEnabled;

    ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
    I(item);

    // no special flags for pseudo items
    PseudoItem * psitem = dynamic_cast<PseudoItem *>(item);
    if (psitem)
    {
        return flags;
    }

    InventoryItem * invitem = dynamic_cast<InventoryItem *>(item);
    if (invitem)
    {
        /*
        // Disabled until we have figured out how to implement renaming properly
        if (item->isNewNode())
        {
            flags |= Qt::ItemIsEditable | Qt::ItemIsDragEnabled;

            if (item->isDirectory())
            {
                flags |= Qt::ItemIsDropEnabled;
            }
        }
        */
    }
    return flags;
}

// the root item needs to store the header for each field
QVariant Inventory::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        return rootItem->data(section, role);
    }

    return QVariant();
}

QModelIndex Inventory::parent(const QModelIndex & index) const
{
    if (!index.isValid())
    {
        return QModelIndex();
    }

    ModelItem * childItem = static_cast<ModelItem *>(index.internalPointer());
    I(childItem);
    ModelItem * parentItem = childItem->parent();

    if (parentItem == rootItem)
    {
        return QModelIndex();
    }

    I(parentItem);
    return createIndex(parentItem->row(), 0, parentItem);
}

int Inventory::rowCount(const QModelIndex & parent) const
{
    ModelItem * parentItem = rootItem;

    if (parent.isValid())
    {
        parentItem = static_cast<ModelItem *>(parent.internalPointer());
    }

    I(parentItem);
    return parentItem->childCount();
}

//! TODO move this somewhere else without the inventory model dependency
QMap<FileEntry, FileEntryList> Inventory::findUnaccountedRenames()
{
    QList<InventoryItem *> missingItems;
    QList<InventoryItem *> unknownItems;

    foreach (InventoryItem * item, itemMap.values())
    {
        if (item->hasStatus(InventoryItem::Missing))
        {
            missingItems.append(item);
        }

        if (item->hasStatus(InventoryItem::Unknown))
        {
            unknownItems.append(item);
        }
    }

    // TODO: progess bar here!

    QString parentRev = MonotoneUtil::getBaseWorkspaceRevision(workspacePath);
    FileEntryList parentList =  MonotoneUtil::getRevisionManifest(
        MonotoneUtil::getDatabaseFile(workspacePath),
        parentRev
    );

    QMap<FileEntry, FileEntryList> unaccountedRenames;
    QMap<QString, QString> fileIds;

    foreach (InventoryItem * missingItem, missingItems)
    {
        FileEntryList candidates;
        FileEntry entry;

        // this is a new entry not recorded in the base roster
        if (missingItem->hasStatus(InventoryItem::Added) ||
            missingItem->hasStatus(InventoryItem::RenameTarget))
        {
            entry.path = missingItem->getPath();
            entry.is_dir = missingItem->isDirectory();
        }
        else
        {
            bool found = false;
            foreach (entry, parentList)
            {
                if (entry.path == missingItem->getPath())
                {
                    found = true;
                    break;
                }
            }
            I(found);
        }

        foreach (InventoryItem * unknownItem, unknownItems)
        {
            QString unknownPath = unknownItem->getPath();

            // calculate the file id of the unknown file
            if (!unknownItem->isDirectory() &&
                !fileIds.contains(unknownPath))
            {
                QString fileid = MonotoneUtil::getFileId(
                    MonotoneUtil::getDatabaseFile(workspacePath),
                    unknownPath
                );
                // file was not readable, etc.
                if (fileid.isEmpty()) continue;
                fileIds.insert(unknownPath, fileid);
            }

            // at first do a simple file name check
            if ((missingItem->getFilename() == unknownItem->getFilename()) &&
                (missingItem->isDirectory() == unknownItem->isDirectory()))
            {
                FileEntry can(unknownPath, unknownItem->isDirectory());
                if (fileIds.contains(unknownPath))
                {
                    can.fileid = fileIds.value(unknownPath);
                }
                candidates.append(can);
                continue;
            }

            // we can't do anything for directories from here on
            if (missingItem->isDirectory() || unknownItem->isDirectory())
                continue;

            // we now rely on the fact that we have a fileid
            I(fileIds.contains(unknownPath));

            if (fileIds.value(unknownPath) == entry.fileid)
            {
                candidates.append(FileEntry(unknownPath, false, entry.fileid));
            }
        }

        // only add those missing items to the map which are having any
        // candidates, i.e. are likely to be renamed
        if (candidates.size() > 0)
        {
            unaccountedRenames.insert(entry, candidates);
        }
    }

    return unaccountedRenames;
}

void Inventory::insertRowsRecursive(ModelItem * parentItem, const QMap<ModelItem *, QList<ModelItem *> > & parentChildRelations)
{
    I(parentChildRelations.contains(parentItem));
    QModelIndex parentIndex = indexFromItem(parentItem, 0);
    I(parentIndex.isValid() || parentItem->isRoot());
    const QList<ModelItem *> & children = parentChildRelations.value(parentItem);
    I(children.size() > 0);

    //
    // ensure that old rows get deleted at first
    //
    foreach (ModelItem * item, children)
    {
        InventoryItem * invitem = dynamic_cast<InventoryItem*>(item);
        if (!invitem) continue;

        QString path = invitem->getPath();
        if (itemMap.contains(path))
        {
            InventoryItem * oldItem = itemMap.value(path);
            ModelItem * oldParent = oldItem->parent();
            QModelIndex oldParentIndex = indexFromItem(oldParent, 0);
            int row = oldItem->row();
            removeRowsRecursive(oldParentIndex, row, row);
        }
    }

    //
    // now do the actual row insertion while notifying Qt properly about that
    //
    int parentRowCount = rowCount(parentIndex);
    beginInsertRows(parentIndex, parentRowCount, parentRowCount + children.size() - 1);
    foreach (ModelItem * item, children)
    {
        InventoryItem * invitem = dynamic_cast<InventoryItem*>(item);
        if (invitem)
        {
            QString path = invitem->getPath();
            I(!itemMap.contains(path));
            itemMap.insert(path, invitem);
        }
        parentItem->appendChild(item);
    }
    endInsertRows();

    //
    // finally check if the just added items have children we need
    // to insert recursively
    //
    foreach (ModelItem * item, children)
    {
        if (parentChildRelations.contains(item))
            insertRowsRecursive(item, parentChildRelations);
    }
}

void Inventory::removeRowsRecursive(const QModelIndex & parent, int startRow, int endRow)
{
    I(startRow >= 0);
    I(parent.isValid() || parent.parent() == QModelIndex());
    I(endRow >= startRow && endRow < rowCount(parent));

    // we need to ensure that we recursively delete all childrens before we
    // actually delete the requested items for parent, otherwise Qt will
    // throw too many signals around
    QModelIndexList indexesToDelete;
    for (int i=startRow; i<=endRow; i++)
    {
        QModelIndex child = index(i, 0, parent);
        I(child.isValid());
        int count = rowCount(child);
        if (count > 0)
        {
            removeRowsRecursive(child, 0, count - 1);
        }

        indexesToDelete.append(child);
    }

    beginRemoveRows(parent, startRow, endRow);
    foreach (QModelIndex index, indexesToDelete)
    {
        I(index.isValid());
        I(rowCount(index) == 0);

        ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
        I(item);

        InventoryItem * invitem = qobject_cast<InventoryItem *>(item);
        if (invitem)
        {
            itemMap.remove(invitem->getPath());
        }

        item->parent()->removeChild(item);
        delete item;
    }
    endRemoveRows();
}

