/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "MonotoneManager.h"
#include "MonotoneUtil.h"
#include "vocab.h"
#include "BasicIOParser.h"
#include "SignalWaiter.h"

#include <QMetaType>
#include <QMessageBox>

const QString MonotoneManager::MinInterfaceVersion = MIN_MTN_INT_VERSION;
const QString MonotoneManager::MaxInterfaceVersion = MAX_MTN_INT_VERSION;

MonotoneManager::MonotoneManager() : threadNumber(0) {}

MonotoneManager::~MonotoneManager()
{
    QMapIterator<int, MonotoneThread *> it(threadMap);
    while(it.hasNext())
    {
        it.next();
        int threadNumber = it.key();
        MonotoneThread * thread = it.value();
        D(QString("aborting thread %1").arg(threadNumber));
        thread->abort();
        thread->wait();
        threadMap.remove(threadNumber);
        delete thread;
    }
}

bool MonotoneManager::setMonotoneBinaryPath(const QString & mtnBinary,
    bool relaxedVersionCheck, QString & pathVersion)
{
    pathVersion = getInterfaceVersion(mtnBinary);

    // check if we match the minimum interface version
    if (versionCompare(pathVersion, MinInterfaceVersion) < 0)
    {
        return false;
    }

    // check if we should check and match also the maximum interface version
    if (!relaxedVersionCheck && versionCompare(pathVersion, MaxInterfaceVersion) >= 0)
    {
        return false;
    }

    mtnPath = mtnBinary;
    return true;
}

// FIXME: I think we need to care somehow if we pass
// threads around - i.e. use shared ptrs or something
MonotoneThread * MonotoneManager::getThreadForWorkspace(const WorkspacePath & workspace)
{
    // ensure that its normalized
    WorkspacePath normalizedWorkspace = normalizeWorkspacePath(workspace);
    DatabaseFile databaseFile = getDatabaseFilePath(normalizedWorkspace);
    return getThread(databaseFile, normalizedWorkspace);
}

MonotoneThread * MonotoneManager::getThreadForDatabase(const DatabaseFile & database)
{
    QFile dbfile(database);
    if (!dbfile.open(QIODevice::ReadOnly))
    {
        throw GuitoneException(tr("Could not open database for reading"));
    }

    if (dbfile.read(15) != QByteArray("SQLite format 3"))
    {
        dbfile.close();
        throw GuitoneException(tr("Database file not an SQLite 3 database"));
    }

    if (!dbfile.seek(60) || dbfile.read(4) != QByteArray("_MTN"))
    {
        dbfile.close();
        throw GuitoneException(tr("SQLite database is not a monotone "
                                  "database or a database of an older, "
                                  "unsupported version of monotone"));
    }
    dbfile.close();
    return getThread(database, QString());
}

// FIXME: if there is already a thread "path/to/db.mtn|path/to/workspace" running,
// re-use this if we get at request for "path/to/db.mtn|"
MonotoneThread * MonotoneManager::getThread(const DatabaseFile & database, const WorkspacePath & workspace)
{
    QMutexLocker locker(&lock);

    if (mtnPath.isEmpty())
        throw GuitoneException(tr("Monotone binary path not set"));

    // since we cannot set the workspace directory after we've
    // started the process, we need to ensure that each workspace runs with
    // its own process
    QString ident = database + "|" + workspace;
    QList<int> threadIDs = identMap.keys(ident);

    if (threadIDs.size() == 0)
    {
        threadNumber++;

        MonotoneThread * thread =
            new MonotoneThread(threadNumber, mtnPath, database, workspace);

        threadMap.insert(threadNumber, thread);
        identMap.insert(threadNumber, ident);
        threadIDs.append(threadNumber);

        static bool processErrorRegisteredMetaType = false;
        if (!processErrorRegisteredMetaType)
        {
            qRegisterMetaType<QProcess::ProcessError>("QProcess::ProcessError");
            processErrorRegisteredMetaType = true;
        }

        connect(
            thread, SIGNAL(aborted(int, QProcess::ProcessError, const QString &)),
            this, SLOT(aborted(int, QProcess::ProcessError, const QString &))
        );

        thread->start();
    }

    // TODO: we may want to add support for multiple threads for one
    // and the same database here in the future... then we should probably
    // also make a smarter choice of available threads here instead of just
    // taking the first!
    return threadMap.value(threadIDs.value(0));
}

WorkspacePath MonotoneManager::normalizeWorkspacePath(const WorkspacePath & workspace)
{
    QDir tempDir(workspace);
    if (!tempDir.exists())
    {
        throw GuitoneException(tr("Directory %1 does not exist").arg(workspace));
    }

    bool found = false;
    WorkspacePath normalizedWorkspace;
    do
    {
        if (tempDir.cd("_MTN"))
        {
            tempDir.cdUp();
            normalizedWorkspace = tempDir.absolutePath();
            found = true;
            break;
        }
    }
    while (!tempDir.isRoot() && tempDir.cdUp());

    if (!found)
    {
        throw GuitoneException(tr("could not find _MTN directory"));
    }

    return normalizedWorkspace;
}

// we assume that the workspace was already normalized here
DatabaseFile MonotoneManager::getDatabaseFilePath(const WorkspacePath & workspace)
{
    // now check again if we know it
    if (workspaceMap.contains(workspace))
    {
        return workspaceMap.value(workspace);
    }

    QFile optionsFile(workspace + "/_MTN/options");
    if (!optionsFile.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        throw GuitoneException(tr("could not open _MTN/options for reading"));
    }

    QByteArray contents = optionsFile.readAll();
    optionsFile.close();

    if (contents.size() == 0)
    {
        throw GuitoneException(tr("file _MTN/options is empty"));
    }

    BasicIOParser parser(QString::fromUtf8(contents));
    if (!parser.parse())
    {
        throw GuitoneException(tr("could not parse basic_io from _MTN/options"));
    }
    StanzaList stanzas = parser.getStanzas();
    I(stanzas.size() == 1);
    Stanza st = stanzas.at(0);

    DatabaseFile databaseFilePath;
    foreach (StanzaEntry entry, st)
    {
        if (entry.sym == "database")
        {
            I(entry.vals.size() == 1);
            databaseFilePath = entry.vals.at(0);
            break;
        }
    }

    if (databaseFilePath.isEmpty())
    {
        throw GuitoneException(tr("could not find database for workspace"));
    }

    // remember what we've just found for later requests
    workspaceMap.insert(workspace, databaseFilePath);
    return databaseFilePath;
}

void MonotoneManager::aborted(int threadNumber, QProcess::ProcessError error, const QString & message)
{
    QMutexLocker locker(&lock);

    I(threadMap.contains(threadNumber));

    MonotoneThread * thread = threadMap.value(threadNumber);

    QString databaseFile  = thread->getDatabaseFilePath();
    QString workspacePath = thread->getWorkspacePath();

    disconnect(
        thread, SIGNAL(aborted(int, QProcess::ProcessError, const QString &)),
        this, SLOT(aborted(int, QProcess::ProcessError, const QString &))
    );

    threadMap.remove(threadNumber);
    identMap.remove(threadNumber);

    delete thread;

    //
    // the thread has been removed cleanly, now report what was going on
    //

    QString mtnError = MonotoneUtil::stripMtnPrefix(message);
    QString processErrorTranslated;

    switch (error)
    {
        case QProcess::FailedToStart:
            processErrorTranslated = tr("failed to start");
            break;
        case QProcess::Crashed:
            processErrorTranslated = tr("crashed");
            break;
        case QProcess::Timedout:
            processErrorTranslated = tr("timed out waiting for data");
            break;
        case QProcess::WriteError:
            processErrorTranslated = tr("failed to read");
            break;
        case QProcess::ReadError:
            processErrorTranslated = tr("failed to write");
            break;
        case QProcess::UnknownError:
            processErrorTranslated = tr("unknown error");
            break;
        default: I(false);
    }

    // log something reasonable
    C(QString("thread %1 died (%2) - stderr was %3")
        .arg(threadNumber).arg(processErrorTranslated).arg(mtnError));

    // display something understandable to the user
    QString userMessage(tr(
        "The monotone process for '%1' died (%2). "
        "If you think this is a bug in guitone, "
        "please report it to the author!"
    ).arg(workspacePath.isEmpty() ? databaseFile : workspacePath)
     .arg(processErrorTranslated)
    );

    if (!mtnError.isEmpty())
    {
        userMessage.append(tr("\n\nLast error output was:\n%1").arg(mtnError));
    }

    QMessageBox::critical(
        NULL,
        tr("Error"),
        userMessage,
        QMessageBox::Ok, 0, 0
    );
}

bool MonotoneManager::singleRun(const QString & mtnBinary,
                                const QStringList & params, QByteArray & output)
{
    QProcess proc;
    proc.start(mtnBinary, params);

    // could not be started (invalid path, not executable, ...)
    if (!proc.waitForStarted(5000))
    {
        C(QString("`%1 %2` not started").arg(mtnBinary).arg(params.join(" ")));
        return false;
    }

    // process doesn't return, etc...
    if (!proc.waitForFinished(5000))
    {
        C(QString("`%1 %2` not finished").arg(mtnBinary).arg(params.join(" ")));
        return false;
    }

    // read all of the output
    output.append(proc.readAll());
    return true;
}

QString MonotoneManager::getInterfaceVersion(const QString & mtnBinary)
{
    QByteArray output;

    if (!singleRun(mtnBinary, QStringList() << "automate" << "interface_version", output))
    {
        C("Couldn't execute `automate interface_version`");
        return QString();
    }

    QString versionString(QString::fromUtf8(output));
    QRegExp regex("^(\\d+(?:\\.\\d+))");

    if (regex.indexIn(versionString) == -1)
    {
        D(QString("couldn't parse output: %1").arg(versionString));
        return false;
    }

    return regex.cap(1);
}

/*!
    Returns
    -1 if right is greater
    0 if both are equal
    1 if left is greater
    \sa checkInterfaceVersion()
*/
int MonotoneManager::versionCompare(const QString & left, const QString & right)
{
    QStringList leftParts = left.split(".");
    QStringList rightParts = right.split(".");

    int leftCount = leftParts.size();
    int rightCount = rightParts.size();
    int maxCount = leftCount > rightCount ? leftCount : rightCount;

    for (int i=0, j=maxCount; i<j; i++)
    {
        unsigned int l = 0;
        if (i < leftCount) { l = leftParts.at(i).toUInt(); }
        unsigned int r = 0;
        if (i < rightCount) { r = rightParts.at(i).toUInt(); }

        if (l == r) continue;
        if (l > r) return 1;
        return -1;
    }

    return 0;
}
