/***************************************************************************
*   Copyright (C) 2006 by Thomas Keller                                   *
*   me@thomaskeller.biz                                                   *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "DiffParser.h"
#include "vocab.h"

#include <QStringList>
#include <QRegExp>

DiffParser::DiffParser(const QString & input)
{
    parse(input);
}

DiffParser::~DiffParser()
{
    qDeleteAll(fileDiffs);
    fileDiffs.clear();
}

void DiffParser::parse(const QString & input)
{
    if (input.length() == 0)
    {
        W("Nothing to parse");
        return;
    }

    // remove last newline character
    QString in(input);
    in.chop(1);

    // split the output into lines
    QStringList lines = in.split(QChar('\n'));

    QString curFile;

    Diff * curDiff = 0;
    DiffHunk * curHunk = 0;

    for (int i=0, s=lines.size(); i<s; i++)
    {
        QString line(lines.at(i));
        I(line.length() > 0);
        QChar curChar = line.at(0);

        //
        // parse file diff separators, setup new DiffFile object
        //
        if (curChar == '=')
        {
            I(i < (s-1));

            // straightly go to the next line
            QString nextGroup(lines.at(++i));

            // create a new diff
            curDiff = new Diff();

            // check if this is a binary
            QRegExp rx = QRegExp("^#\\s(.+)\\sis binary");
            if (rx.indexIn(nextGroup) != -1)
            {
                curDiff->is_binary = true;
                curFile = rx.cap(1);
            }
            else
            {
                // then this should be a normal file diff
                rx = QRegExp("^---\\s([^\\t]+)\\t\\w{40}");
                I(rx.indexIn(nextGroup) > -1);
                curFile = rx.cap(1);
                curDiff->is_binary = false;
                // skip the next line as we don't care about renames
                // yet and if we care about them, we get this info from
                // the cset
                i++;
            }

            fileDiffs.insert(curFile, curDiff);
            continue;
        }

        I(curDiff);

        //
        // parse hunk lines, setup new hunk object
        //
        if (curChar == '@')
        {
            QRegExp rx = QRegExp("^@@\\s+\\-(\\d+)(,(\\d+))?\\s+\\+(\\d+)(,(\\d+))?\\s+@@");
            I(rx.indexIn(line) > -1);
            QStringList list = rx.capturedTexts();
            I(list.size() >= 2);

            // create a new Hunk
            curHunk = new DiffHunk();
            curDiff->hunks.push_back(curHunk);

            // extract line numbers and positions
            curHunk->leftStart = list.at(1).toInt();
            if (list.at(2).length() > 0)
                curHunk->leftCount = list.at(3).toInt();
            else
                curHunk->leftCount = 1;

            curHunk->rightStart = list.at(4).toInt();
            if (list.at(5).length() > 0)
                curHunk->rightCount = list.at(6).toInt();
            else
                curHunk->rightCount = 1;

            continue;
        }

        I(curHunk);

        DiffLine * diffLine = new DiffLine();
        curHunk->lines.push_back(diffLine);
        diffLine->content = line.mid(1);

        if (curChar == '+') diffLine->state = DiffLine::Added;
        else if (curChar == '-') diffLine->state = DiffLine::Removed;
        else if (curChar == ' ') diffLine->state = DiffLine::Unchanged;
        else I(false);
    }
}

Diff * DiffParser::getDiff(const QString & filename)
{
    if (!fileDiffs.contains(filename)) return 0;
    return fileDiffs.value(filename);
}

