/**************************************************************************
*   Copyright (C) 2006 by Jean-Louis Fuchs                                *
*   ganwell@fangorn.ch                                                    *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "TreeBuilder.h"
#include <stdlib.h>
#include <time.h>

TreeBuilder::TreeBuilder(QStandardItem * r, QObject * parent)
    : QObject(parent), root(r), sep('.') {}

TreeBuilder::~TreeBuilder() {};

QStandardItem * TreeBuilder::add(const QString & branch)
{
    if (branch != "")
    {
        return add(branch, root);
    }
    return NULL;
}

void TreeBuilder::addData(QStandardItem * child)
{
    QString branch;
    QStandardItem * item = child;
    QString text = child->data().toString();

    if (text.right(1) == QString(sep))
    {
        text = text.left(text.length() - 1) + "*";
    }
    child->setText(text);

    while (item)
    {
        branch = item->data().toString() + branch;
        item = item->parent();
    }

    if (branch.right(1) == QString(sep))
    {
        branch = branch.left(branch.length() - 1) + "*";
    }

    child->setData(branch, Qt::ToolTipRole);
}

QStandardItem * TreeBuilder::add(const QString & branch, QStandardItem * parent)
{
    if (branch == "")
    {
        return NULL;
    }

    int len = parent->rowCount();
    for(int index = 0; index < len; index++)
    {
        int pos = overlap(parent->child(index)->data().toString(), branch) + 1;
        if (pos)
        {
            // Store the child data as we need multiple times
            QString tmp = parent->child(index)->data().toString();

            if (pos == tmp.length())
            {
                // The child contains the branch.
                return add(branch.mid(pos), parent->child(index));
            }
            else
            {
                // We found a splitting point. Going to splitt the tree.

                // Get the old childs rows
                QList<QStandardItem *> row = parent->child(index)->takeColumn(0);

                // Delete the old child
                parent->removeRow(index);

                // Create a new parent to replace the old child
                QStandardItem * newparent = new QStandardItem();
                newparent->setData(branch.left(pos));

                // Add the new parent to the parent
                parent->appendRow(newparent);
                addData(newparent);

                // Add the old child to the new parent
                QStandardItem * ret;
                ret = add(tmp.mid(pos), newparent);

                // Append the old childs rows to the newly added old child
                QListIterator<QStandardItem *> iterator(row);
                while (iterator.hasNext())
                {
                    ret->appendRow(iterator.next());
                }

                // Add the new child to the new parent
                add(branch.mid(pos), newparent);

                return newparent;
            }
        }
    }

    QStandardItem * it = new QStandardItem();
    it->setData(branch);
    parent->appendRow(it);
    addData(it);
    return it;
}

int TreeBuilder::overlap(const QString & a, const QString & b)
{
    const QChar * ac = a.data();
    const QChar * bc = b.data();
    int len = 0;
    int max = qMax(a.length(), b.length());

    while (len < max && ac[len] == bc[len])
    {
        len++;
    }

    // Important: It only is a splitting point if both a and b have a sep at
    // this point!
    while (len >= 0 && (ac[len] != sep || bc[len] != sep))
    {
        len--;
    }
    return len;
}

void TreeBuilder::addList(const QString & branches)
{
    QStringList branchList = branches.split("\n", QString::SkipEmptyParts);

    foreach (QString branch, branchList)
    {
        add(branch);
    }
}
