/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "InventoryView.h"
#include "Settings.h"
#include "InventoryItem.h"
#include "vocab.h"

#include <QMenu>
#include <QHeaderView>
#include <QSortFilterProxyModel>
#include <QMessageBox>
#include <QTimer>

InventoryView::InventoryView(QWidget * parent, const QString & objectName)
  : TreeView(parent, objectName), invViewDelegate(parent)
{
    setItemDelegate(&invViewDelegate);
    setSelectionMode(QAbstractItemView::ExtendedSelection);
    setDragEnabled(true);
    setDropIndicatorShown(true);
    setSortingEnabled(true);
    sortByColumn(0, Qt::AscendingOrder);

    header()->setClickable(true);

    createAndConnectContextActions();

    connect(
        this, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(slotContextMenuRequested(const QModelIndexList &, const QPoint &))
    );
}

void InventoryView::setType(Type t)
{
    // ensure that a model is already set, otherwise the app crashes
    I(model());

    type = t;

    if (type == FileList)
    {
        disconnect(
            this, SIGNAL(clicked(const QModelIndex &)),
            this, SLOT(itemClicked(const QModelIndex &))
        );

        connect(
            this, SIGNAL(doubleClicked(const QModelIndex &)),
            this, SLOT(itemDoubleClicked(const QModelIndex &))
        );

        setRootIsDecorated(false);
        setItemsExpandable(false);
    }
    else
    {
        disconnect(
            this, SIGNAL(clicked(const QModelIndex &)),
            this, SLOT(itemDoubleClicked(const QModelIndex &))
        );

        connect(
            this, SIGNAL(clicked(const QModelIndex &)),
            this, SLOT(itemClicked(const QModelIndex &))
        );
        setRootIsDecorated(true);
        setItemsExpandable(true);
    }
}

void InventoryView::createAndConnectContextActions()
{
    actChdir = new QAction(tr("Go into"), this);
    actChdir->setStatusTip(tr("Go into the directory"));
    connect(actChdir, SIGNAL(triggered()), this, SLOT(slotChdir()));

    actOpen = new QAction(tr("Open"), this);
    actOpen->setStatusTip(tr("Open in default program"));
    connect(actOpen, SIGNAL(triggered()), this, SLOT(slotOpen()));

    actAdd = new QAction(tr("Add"), this);
    actAdd->setStatusTip(tr("Add to workspace"));
    connect(actAdd, SIGNAL(triggered()), this, SLOT(slotAdd()));

    actRemove = new QAction(tr("Remove"), this);
    actRemove->setStatusTip(tr("Remove from workspace"));
    connect(actRemove, SIGNAL(triggered()), this, SLOT(slotRemove()));

    actCommit = new QAction(tr("Commit"), this);
    actCommit->setStatusTip(tr("Commit"));
    connect(actCommit, SIGNAL(triggered()), this, SLOT(slotCommit()));

    actIgnore = new QAction(tr("Ignore"), this);
    actIgnore->setStatusTip(tr("Ignore file"));
    connect(actIgnore, SIGNAL(triggered()), this, SLOT(slotIgnore()));

    actUnignore = new QAction(tr("Unignore"), this);
    actUnignore->setStatusTip(tr("Unignore file"));
    connect(actUnignore, SIGNAL(triggered()), this, SLOT(slotUnignore()));

    actRevert = new QAction(tr("Revert"), this);
    actRevert->setStatusTip(tr("Revert uncommitted changes"));
    connect(actRevert, SIGNAL(triggered()), this, SLOT(slotRevert()));

    actFileDiff = new QAction(tr("Diff"), this);
    actFileDiff->setStatusTip(tr("Diff against base revision"));
    connect(actFileDiff, SIGNAL(triggered()), this, SLOT(slotFileDiff()));

    actFileHistory = new QAction(tr("History"), this);
    actFileHistory->setStatusTip(tr("Display the history of this file"));
    connect(actFileHistory, SIGNAL(triggered()), this, SLOT(slotFileHistory()));

    actRevisionDiff = new QAction(tr("Diff all"), this);
    actRevisionDiff->setStatusTip(tr("Show all differences"));
    connect(actRevisionDiff, SIGNAL(triggered()), this, SLOT(slotRevisionDiff()));

    actRename = new QAction(tr("Rename"), this);
    actRename->setStatusTip(tr("Rename file"));
    connect(actRename, SIGNAL(triggered()), this, SLOT(slotRename()));

    actAddMultiple = new QAction(tr("Add %1 items"), this);
    actAddMultiple->setStatusTip(tr("Add multiple items"));
    connect(actAddMultiple, SIGNAL(triggered()), this, SLOT(slotAdd()));

    actRemoveMultiple = new QAction(tr("Remove %1 items"), this);
    actRemoveMultiple->setStatusTip(tr("Remove multiple items"));
    connect(actRemoveMultiple, SIGNAL(triggered()), this, SLOT(slotRemove()));

    actCommitMultiple = new QAction(tr("Commit %1 items"), this);
    actCommitMultiple->setStatusTip(tr("Commit multiple items"));
    connect(actCommitMultiple, SIGNAL(triggered()), this, SLOT(slotCommit()));

    actIgnoreMultiple = new QAction(tr("Ignore %1 items"), this);
    actIgnoreMultiple->setStatusTip(tr("Ignore multiple items"));
    connect(actIgnoreMultiple, SIGNAL(triggered()), this, SLOT(slotIgnore()));

    actUnignoreMultiple = new QAction(tr("Unignore %1 items"), this);
    actUnignoreMultiple->setStatusTip(tr("Unignore multiple items"));
    connect(actUnignoreMultiple, SIGNAL(triggered()), this, SLOT(slotUnignore()));

    actRevertMultiple = new QAction(tr("Revert %1 items"), this);
    actRevertMultiple->setStatusTip(tr("Revert multiple items"));
    connect(actRevertMultiple, SIGNAL(triggered()), this, SLOT(slotRevert()));
}

InventoryView::~InventoryView()
{
    delete actChdir;
    delete actOpen;
    delete actAdd;
    delete actRemove;
    delete actCommit;
    delete actIgnore;
    delete actUnignore;
    delete actRevert;
    delete actFileDiff;
    delete actFileHistory;
    delete actRevisionDiff;
    delete actRename;
    delete actAddMultiple;
    delete actRemoveMultiple;
    delete actCommitMultiple;
    delete actIgnoreMultiple;
    delete actUnignoreMultiple;
    delete actRevertMultiple;
}

// ensure that we only accept sortfilterproxymodels
void InventoryView::setModel(QAbstractItemModel * model)
{
    Q_UNUSED(model);
    I(false);
}

void InventoryView::setModel(QSortFilterProxyModel * newModel)
{
    QAbstractItemModel * oldModel = model();
    if (oldModel)
    {
        disconnect(
            oldModel, SIGNAL(dataChanged(const QModelIndex &, const QModelIndex &)),
            this, SLOT(newNodeLoaded(const QModelIndex &, const QModelIndex &))
        );

        disconnect(
            oldModel, SIGNAL(rowsInserted(const QModelIndex &, int, int)),
            this, SLOT(setBackgroundImage())
        );

        disconnect(
            oldModel, SIGNAL(rowsRemoved(const QModelIndex &, int, int)),
            this, SLOT(setBackgroundImage())
        );

        disconnect(
            oldModel, SIGNAL(modelReset()),
            this, SLOT(setBackgroundImage())
        );
    }

    TreeView::setModel(newModel);

    connect(
        newModel, SIGNAL(dataChanged(const QModelIndex &, const QModelIndex &)),
        this, SLOT(newNodeLoaded(const QModelIndex &, const QModelIndex &))
    );

    connect(
        newModel, SIGNAL(rowsInserted(const QModelIndex &, int, int)),
        this, SLOT(setBackgroundImage())
    );

    connect(
        newModel, SIGNAL(rowsRemoved(const QModelIndex &, int, int)),
        this, SLOT(setBackgroundImage())
    );

    connect(
        newModel, SIGNAL(modelReset()),
        this, SLOT(setBackgroundImage())
    );
}

void InventoryView::newNodeLoaded(const QModelIndex & index, const QModelIndex & dummy)
{
    Q_UNUSED(dummy);

    if (!index.isValid())
    {
        C("Index invalid, aborting...");
        return;
    }

    if (type == FileList)
    {
        setRootIndex(index);
    }
    else
    {
        expand(index);
    }
}

// adds elements to the popup menu based on the selected items
// since not all actions may apply on all items
void InventoryView::slotContextMenuRequested(const QModelIndexList & indexList, const QPoint & pos)
{
    if (indexList.size() == 0) return;

    //
    // single item selected
    //
    if (indexList.size() == 1)
    {
        QModelIndex sourceIndex = static_cast<const QSortFilterProxyModel*>
            (indexList[0].model())->mapToSource(indexList[0]);
        ModelItem * item = static_cast<ModelItem *>
            (sourceIndex.internalPointer());
        InventoryItem * invitem = dynamic_cast<InventoryItem *>
            (item);

        // only display the popup for real workspace items
        if (!invitem)
            return;

        QMenu menu(this);

        if (invitem->isDirectory())
        {
            menu.addAction(actChdir);
        }

        if (invitem->hasStatus(InventoryItem::Unknown) ||
            invitem->hasStatus(InventoryItem::Ignored) || (
               invitem->isNewNode() &&
               invitem->hasNotStatus(InventoryItem::Missing) &&
               invitem->hasNotStatus(InventoryItem::Invalid)
           ))
        {
            menu.addAction(actOpen);
        }

        if (invitem->hasStatus(InventoryItem::Unknown))
        {
            menu.addAction(actAdd);
            menu.addAction(actIgnore);
        }

        if (invitem->hasStatus(InventoryItem::Ignored))
        {
            menu.addAction(actUnignore);
        }

        if (invitem->hasChangedRecursive())
        {
            menu.addAction(actCommit);

            if (invitem->isDirectory() &&
                invitem->getFSType() == InventoryItem::Directory)
            {
                menu.addAction(actRevisionDiff);
            }
        }

        // FIXME: we currently exclude the workspace root for this.
        // While its reasonable for the remove action, the root could
        // actually be renamed by pivot_root - decide whether we want
        // to implement that here or offer a workspace menu and dialog
        // for this rather rare use case
        if (invitem->hasNotStatus(InventoryItem::Ignored) &&
            invitem->hasNotStatus(InventoryItem::Unknown) &&
            !invitem->getPath().isEmpty())
        {
            menu.addAction(actRemove);
            menu.addAction(actRename);

            // as long as it is not unchanged, it can be converted
            if (invitem->hasChanged())
            {
                menu.addAction(actRevert);

                if (invitem->hasStatus(InventoryItem::ContentsChanged) &&
                    invitem->hasNotStatus(InventoryItem::Added))
                {
                    menu.addAction(actFileDiff);
                }
            }

            if (invitem->getFSType() == InventoryItem::File &&
                invitem->hasNotStatus(InventoryItem::Added))
            {
                menu.addAction(actFileHistory);
            }
        }

        //
        // mark the default action with bold text
        //
        DefaultAction act = getDefaultAction(indexList[0]);
        QFont activeFont;
        activeFont.setBold(true);
        switch (act)
        {
            case Chdir:
                actChdir->setFont(activeFont);
                break;
            case FileDiff:
                actFileDiff->setFont(activeFont);
                break;
            case Commit:
                actCommit->setFont(activeFont);
                break;
            case Open:
                actOpen->setFont(activeFont);
                break;
            default:
                break;
        }

        menu.exec(pos);
        return;
    }

    //
    // multiple items selected
    //
    ModelItem * item;
    InventoryItem * invitem;
    QModelIndex sourceIndex;

    QStringList actions;
    actions << "add" << "drop" << "commit"
            << "ignore" << "unignore" << "revert";

    QMap<QString, QAction *> actionMapping;
    actionMapping.insert("add", actAddMultiple);
    actionMapping.insert("drop", actRemoveMultiple);
    actionMapping.insert("commit", actCommitMultiple);
    actionMapping.insert("revert", actRevertMultiple);
    actionMapping.insert("ignore", actIgnoreMultiple);
    actionMapping.insert("unignore", actUnignoreMultiple);

    QMap<QString, int> actionCounter;

    foreach (QString action, actions)
    {
        actionCounter.insert(action, 0);
    }

    foreach (QModelIndex index, indexList)
    {
        sourceIndex =
            static_cast<const QSortFilterProxyModel *>(index.model())
                ->mapToSource(index);

        item = static_cast<ModelItem *>(sourceIndex.internalPointer());
        invitem = dynamic_cast<InventoryItem *>(item);

        // skip actions on anything other than InventoryItems
        if (!invitem)
            continue;

        if (!invitem->isTracked())
        {
            actionCounter["add"]++;

            if (invitem->hasStatus(InventoryItem::Unknown))
            {
                actionCounter["ignore"]++;
            }
            else
            {
                actionCounter["unignore"]++;
            }
            continue;
        }

        if (invitem->hasChangedRecursive())
        {
            actionCounter["commit"]++;
            actionCounter["revert"]++;
        }

        actionCounter["drop"]++;
    }

    QMapIterator<QString, int> i(actionCounter);
    int totalItems = indexList.size();
    bool commonActions = false;

    QMenu menu(this);

    while (i.hasNext())
    {
        i.next();
        if (i.value() < totalItems) continue;
        commonActions = true;
        QAction * act = actionMapping[i.key()];
        act->setText(act->text().arg(totalItems));
        menu.addAction(act);
    }

    if (!commonActions)
    {
        QAction * act = new QAction(
            tr("No common action for selected items"), this
        );
        menu.addAction(act);
    }

    menu.exec(pos);
}

void InventoryView::slotChdir()
{
    QItemSelectionModel * selectionModel = this->selectionModel();
    QList<QModelIndex> list(selectionModel->selectedIndexes());
    I(list.size() > 0);

    emit directoryChanged(list[0]);
    changeDirectory(list[0]);
}

void InventoryView::changeDirectory(const QModelIndex & proxyIndex)
{
    QModelIndex source = dynamic_cast<QSortFilterProxyModel *>
        (model())->mapToSource(proxyIndex);

    if (type == FileList)
    {
        // retranslate the proxy index to an index of the current model
        QModelIndex proxyFileIndex =
            dynamic_cast<QSortFilterProxyModel *>(model())->mapFromSource(source);

        ModelItem * item = static_cast<ModelItem *>
            (source.internalPointer());

        PseudoItem * psitem = dynamic_cast<PseudoItem *>(item);
        if (psitem && psitem->isCdUp())
        {
            // list the contents of the parent of the parent directory
            setRootIndex(proxyFileIndex.parent().parent());
            return;
        }
        setRootIndex(proxyFileIndex);
        return;
    }

    // retranslate the proxy index to an index of the current model
    QModelIndex proxyFolderIndex =
        static_cast<QSortFilterProxyModel *>(model())->mapFromSource(source);

    // expand the selection if needed
    expand(proxyFolderIndex);

    // select the item
    selectionModel()->setCurrentIndex(
        proxyFolderIndex,
        QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows
    );
}

void InventoryView::slotOpen()
{
    QModelIndex index(getSingleSelection());
    if (!index.isValid()) return;

    ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
    InventoryItem * invitem = dynamic_cast<InventoryItem *>(item);

    if (!invitem)
        return;

    emit openFile(invitem->getPath());
}

void InventoryView::slotAdd()
{
    C("Not implemented.");
}

void InventoryView::slotRemove()
{
    C("Not implemented.");
}

void InventoryView::slotCommit()
{
    C("Not implemented.");
}

void InventoryView::slotRevert()
{
    C("Not implemented.");
}

void InventoryView::slotRename()
{
    C("Not implemented.");
}

//
// TODO: there is no mtn API for the next two commands available
// right now so we need to manually edit the WD/.mtn-ignore file
// Q: If this file is under version control, should a change
// be committed in the background?
//
void InventoryView::slotIgnore()
{
    C("Not implemented.");
}

void InventoryView::slotUnignore()
{
    C("Not implemented.");
}

void InventoryView::slotFileDiff()
{
    QModelIndex index(getSingleSelection());
    if (!index.isValid()) return;

    ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
    InventoryItem * invitem = dynamic_cast<InventoryItem *>(item);

    if (!invitem)
        return;

    if (!invitem->hasStatus(InventoryItem::ContentsChanged))
    {
        D("File has no content changes");
        return;
    }

    emit diffFile(invitem->getPath());
}

void InventoryView::slotFileHistory()
{
    QModelIndex index(getSingleSelection());
    if (!index.isValid()) return;

    ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
    InventoryItem * invitem = dynamic_cast<InventoryItem *>(item);

    if (!invitem)
        return;

    QString path = invitem->getPath();

    // determine the original path, if needed
    if (invitem->hasStatus(InventoryItem::RenameSource))
    {
        path = invitem->getRenameTarget();
    }
    if (invitem->hasStatus(InventoryItem::RenameTarget))
    {
        path = invitem->getRenameSource();
    }

    emit fileHistory(path);
}

void InventoryView::slotRevisionDiff()
{
    QModelIndex index(getSingleSelection());
    if (!index.isValid()) return;

    ModelItem * item = static_cast<ModelItem *>(index.internalPointer());
    InventoryItem * invitem = dynamic_cast<InventoryItem *>(item);

    if (!invitem)
        return;

    emit diffRevision(invitem->getPath(), QString(), QString());
}

QModelIndex InventoryView::getSingleSelection(bool mapToSource /* = true */) const
{
    QItemSelectionModel * selectionModel = this->selectionModel();
    QList<QModelIndex> list(selectionModel->selectedIndexes());

    if (list.size() == 0)
    {
        D("No item selected.");
        return QModelIndex();
    }

    if (list.size() > 1)
    {
        D("Multiple items selected, only returning the first.");
    }

    if (!mapToSource) return list[0];

    return static_cast<const QSortFilterProxyModel *>(list[0].model())->mapToSource(list[0]);
}

void InventoryView::itemClicked(const QModelIndex & index)
{
    DefaultAction act = getDefaultAction(index);
    if (act == Chdir)
    {
        emit directoryChanged(index);
        changeDirectory(index);
    }
}

void InventoryView::itemDoubleClicked(const QModelIndex & index)
{
    DefaultAction act = getDefaultAction(index);
    switch (act)
    {
        case Chdir:
            emit directoryChanged(index);
            changeDirectory(index);
            break;
        case FileDiff:
            slotFileDiff();
            break;
        case Commit:
            slotCommit();
            break;
        case Open:
            slotOpen();
            break;
        default:
            break;
    }
}

InventoryView::DefaultAction InventoryView::getDefaultAction(const QModelIndex & index) const
{
    if (!index.isValid())
        return None;

    QModelIndex sourceIndex =
        static_cast<const QSortFilterProxyModel *>(index.model())->mapToSource(index);
    I(sourceIndex.isValid());

    ModelItem * item = static_cast<ModelItem *>(sourceIndex.internalPointer());
    InventoryItem * invitem = dynamic_cast<InventoryItem *>(item);
    PseudoItem * psitem = dynamic_cast<PseudoItem *>(item);

    if ((invitem && invitem->isDirectory()) ||
        (psitem && psitem->isCdUp()))
    {
        return Chdir;
    }

    I(invitem);

    if (invitem->hasStatus(InventoryItem::ContentsChanged) &&
        invitem->hasNotStatus(InventoryItem::Added))
    {
        return FileDiff;
    }

    if (invitem->hasChangedRecursive())
    {
        return Commit;
    }

    if (invitem->hasStatus(InventoryItem::Unknown) ||
        invitem->hasStatus(InventoryItem::Ignored) || (
           invitem->isNewNode() &&
           invitem->hasNotStatus(InventoryItem::Missing) &&
           invitem->hasNotStatus(InventoryItem::Invalid)
       ))
    {
        return Open;
    }

    L(QString("no default action for item %1 applicable")
        .arg(invitem->getLabel()));
    return None;
}

/*!
    If the model contains no data, display a "no matches" background image
    This usually happens when the proxy model has filtered out everything
    and maybe in some weird error case which should not happen at all...
*/
void InventoryView::setBackgroundImage()
{
    if (model()->hasChildren())
    {
        setStyleSheet(QString());
    }
    else
    {
        setStyleSheet("InventoryView { "
                      "background-color: white;"
                      "background-image: url(:/images/no_matches.png);"
                      "background-attachment: fixed;"
                      "background-position: center;"
                      "background-repeat: no-repeat;}");
    }
}
