/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "WorkspaceWindow.h"
#include "Settings.h"
#include "Platform.h"
#include "Guitone.h"
#include "MonotoneUtil.h"
#include "InventoryItem.h"

#include "WorkspaceMenuBar.h"
#include "WorkspaceDialogManager.h"

WorkspaceWindow::WorkspaceWindow() : DatabaseWindow(), mainSplitter(0),
    listSplitter(0), treeView(0), listView(0), attrView(0), statusBar(0),
    workspaceFloater(0),
    invModel(0), attrModel(0), proxyModelFolderTree(0), proxyModelFileList(0)
{
    setObjectName("WorkspaceWindow");
}

WorkspaceWindow::~WorkspaceWindow()
{
    listSplitter->saveState();
    mainSplitter->saveState();

    // order is important here - at first the views inside the splitters,
    // then the splitters themselves!
    if (treeView) delete treeView;
    if (listView) delete listView;
    if (attrView) delete attrView;
    if (listSplitter) delete listSplitter;
    if (mainSplitter) delete mainSplitter;
    if (statusBar) delete statusBar;
    if (workspaceFloater)
    {
        removeDockWidget(workspaceFloater);
        delete workspaceFloater;
    }

    // do the same with the models
    if (proxyModelFolderTree) delete proxyModelFolderTree;
    if (proxyModelFileList) delete proxyModelFileList;
    if (invModel) delete invModel;
    if (attrModel) delete attrModel;
}

void WorkspaceWindow::init()
{
    if (!menuBar)
    {
        menuBar = new WorkspaceMenuBar(this);
    }

    if (!dialogManager)
    {
        dialogManager = new WorkspaceDialogManager(this);
    }

    connect(
        menuBar, SIGNAL(showUpdateWorkspace()),
        dialogManager, SLOT(showUpdateWorkspace())
    );

    connect(
        menuBar, SIGNAL(showCommitRevision()),
        dialogManager, SLOT(showCommitRevision())
    );

    connect(
        menuBar, SIGNAL(showUnaccountedRenames()),
        this, SLOT(checkForUnaccountedRenames())
    );

    DatabaseWindow::init();
}

void WorkspaceWindow::setup()
{
    statusBar = new QStatusBar(this);
    setStatusBar(statusBar);

    mainSplitter = new Splitter(this, "MainSplitter");
    mainSplitter->setOrientation(Qt::Horizontal);

    treeView = new InventoryView(mainSplitter, "FolderTree");
    treeView->setEditTriggers(QAbstractItemView::EditKeyPressed);
    treeView->setSelectionMode(QAbstractItemView::ExtendedSelection);

    listSplitter = new Splitter(mainSplitter, "ListSplitter");
    listSplitter->setOrientation(Qt::Vertical);

    listView = new InventoryView(listSplitter, "FileList");
    listView->setEditTriggers(QAbstractItemView::EditKeyPressed);
    listView->setSelectionMode(QAbstractItemView::ExtendedSelection);

    attrView = new AttributesView(listSplitter, "AttributesView");
    attrView->setAlternatingRowColors(true);

    connect(
        treeView, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(readAttributes(const QModelIndex &))
    );
    connect(
        listView, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(readAttributes(const QModelIndex &))
    );

    connect(
        treeView, SIGNAL(directoryChanged(const QModelIndex &)),
        listView, SLOT(changeDirectory(const QModelIndex &))
    );
    connect(
        listView, SIGNAL(directoryChanged(const QModelIndex &)),
        treeView, SLOT(changeDirectory(const QModelIndex &))
    );

    connect(
        listView, SIGNAL(diffFile(const QString &)),
        dialogManager, SLOT(showFileDiff(const QString &))
    );

    connect(
        listView, SIGNAL(diffRevision(const QString &, const QString &, const QString &)),
        dialogManager, SLOT(showRevisionDiff(const QString &, const QString &, const QString &))
    );

    connect(
        listView, SIGNAL(fileHistory(const QString &)),
        dialogManager, SLOT(showFileHistory(const QString &))
    );

    connect(
        listView, SIGNAL(openFile(const QString &)),
        this, SLOT(openFile(const QString &))
    );

    // the tree view shows only directories, therefor file diff / file history
    // doesn't make sense there
    connect(
        treeView, SIGNAL(diffRevision(const QString &, const QString &, const QString &)),
        dialogManager, SLOT(showRevisionDiff(const QString &, const QString &, const QString &))
    );

    connect(
        treeView, SIGNAL(openFile(const QString &)),
        this, SLOT(openFile(const QString &))
    );

    connect(
        menuBar, SIGNAL(expandTree(bool)),
        treeView, SLOT(expandAllNodes(bool))
    );

    listSplitter->addWidget(listView);
    listSplitter->addWidget(attrView);

    mainSplitter->addWidget(treeView);
    mainSplitter->addWidget(listSplitter);

    setCentralWidget(mainSplitter);

    mainSplitter->restoreState();
    listSplitter->restoreState();

    // models
    invModel = new Inventory(this);
    attrModel = new GetAttributes(this);
    proxyModelFolderTree = new InventoryProxyModel(this, true);
    proxyModelFileList = new InventoryProxyModel(this, false);

    proxyModelFolderTree->setSourceModel(invModel);
    proxyModelFileList->setSourceModel(invModel);

    connect(
        invModel, SIGNAL(invalidWorkspaceFormat(const QString &)),
        this, SLOT(invalidWorkspaceFormat(const QString &))
    );

    connect(
        dialogManager, SIGNAL(revisionCommitted(const QString &)),
        invModel, SLOT(refresh())
    );

    connect(
        menuBar, SIGNAL(reloadWorkspace()),
        invModel, SLOT(refresh())
    );

    connect(
        menuBar, SIGNAL(hideIgnoredFiles(bool)),
        proxyModelFileList, SLOT(setHideIgnoredFiles(bool))
    );

    connect(
        menuBar, SIGNAL(hideIgnoredFiles(bool)),
        proxyModelFolderTree, SLOT(setHideIgnoredFiles(bool))
    );

    connect(
        menuBar, SIGNAL(filterFiles(int)),
        proxyModelFileList, SLOT(setViewOption(int))
    );

    connect(
        menuBar, SIGNAL(filterFiles(int)),
        proxyModelFolderTree, SLOT(setViewOption(int))
    );

    treeView->setModel(proxyModelFolderTree);
    treeView->setType(InventoryView::FolderTree);

    listView->setModel(proxyModelFileList);
    listView->setType(InventoryView::FileList);

    attrView->setModel(attrModel);

    workspaceFloater = new WorkspaceFloater(this);

    dynamic_cast<WorkspaceMenuBar *>(menuBar)
        ->addDockWidgetAction(workspaceFloater->toggleViewAction());
}

void WorkspaceWindow::load(const QString & path)
{
    workspacePath = MonotoneManager::normalizeWorkspacePath(path);
    APP->manager()->getThreadForWorkspace(workspacePath);

    invModel->setWorkspacePath(workspacePath);
    attrModel->setWorkspacePath(workspacePath);

    // read the base set of the inventory
    invModel->refresh();

    // initialize the dialog manager
    reinterpret_cast<WorkspaceDialogManager *>(dialogManager)->init(workspacePath);

    setWindowTitle(
        tr("%1 - workspace mode - guitone").
        arg(MonotoneUtil::getBranchNameShort(workspacePath))
    );

    Settings::addItemToList("RecentWorkspaceList", workspacePath, 5);
}

void WorkspaceWindow::openFile(const QString & filePath)
{
    QFileInfo file(workspacePath + "/" + filePath);
    if (!file.exists())
    {
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("The file you're trying to open does not exist."),
            QMessageBox::Ok, 0, 0
        );
        return;
    }

    if (!Platform::openFile(file.absoluteFilePath()))
    {
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("Unable to open files on your platform - please contact the "
            "author about this problem."),
            QMessageBox::Ok, 0, 0
        );
    }
}

//! FIXME the dialog should query the data itself or retrieve from some global cache:
void WorkspaceWindow::checkForUnaccountedRenames()
{
    QMap<FileEntry, FileEntryList> renames = invModel->findUnaccountedRenames();
    if (renames.size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Nothing found"),
            tr("No unaccounted renames found for this workspace."),
            QMessageBox::Ok
        );
        return;
    }

    reinterpret_cast<WorkspaceDialogManager *>(dialogManager)
        ->showUnaccountedRenames(renames);
}

void WorkspaceWindow::readAttributes(const QModelIndex & index)
{
    QModelIndex sourceIndex = static_cast<const QSortFilterProxyModel *>
        (index.model())->mapToSource(index);
    ModelItem * item = static_cast<ModelItem *>
        (sourceIndex.internalPointer());
    InventoryItem * invitem = dynamic_cast<InventoryItem *>
        (item);

    if (!invitem)
    {
        D("item is not inventory item");
        attrModel->revert();
        return;
    }

    if (invitem->isOldNode() || !invitem->isTracked())
    {
        D("item is old node or not tracked");
        attrModel->revert();
        return;
    }

    attrModel->readAttributes(invitem->getPath());
}

void WorkspaceWindow::invalidWorkspaceFormat(const QString & error)
{
    QMessageBox::critical(
        this,
        tr("Error"),
        tr("Unable to read workspace format:\n%1").arg(error),
        QMessageBox::Ok, 0, 0
    );
    close();
}

