/***************************************************************************
 *   Copyright (C) 2006 by Jean-Louis Fuchs                                 *
 *   ganwell@fangorn.ch                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "ChangesetBrowser.h"
#include "Settings.h"

ChangesetBrowser::ChangesetBrowser(QWidget * parent, const DatabaseFile & db)
  : Dialog(parent), databaseFile(db), branchModel(0)
{
    setupUi(this);
    Dialog::init();

    this->setWindowFlags(this->windowFlags() | Qt::WindowMaximizeButtonHint);
    this->setWindowFlags(this->windowFlags() | Qt::WindowMinimizeButtonHint);

    branchesSplitter->restoreState();
    changesetsSplitter->restoreState();
    changeLogSplitter->restoreState();

    tree = Settings::getBool("ChangesetBrowserTree", false);
    displayBranchesAsTree->setText(
        !tree ? tr("display branches as tree") : tr("display branches flat")
    );

    initTreeWidget();

    connect(
        branches, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(branchesClicked(const QModelIndex &))
    );

    connect(
        displayBranchesAsTree, SIGNAL(clicked()),
        this, SLOT(toggleTree())
    );

    changesetModel = new ChangesetModel(this, databaseFile);
    changesets->setModel(changesetModel);
    changesets->setRootIsDecorated(false);

    revisionModel = new GetRevision(this);
    revisionView->setModel(revisionModel);

    connect(
        changesets, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(changesetsClicked(const QModelIndex &))
    );

    connect(
        pushAll, SIGNAL(clicked()),
        this, SLOT(receiveAll())
    );

    connect(
        pushMore, SIGNAL(clicked()),
        this, SLOT(receiveMore())
    );

    connect(
        changesets, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(changesetsDoubleClicked(const QModelIndex &))
    );

    connect(
        changesets, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(contextMenuRequested(const QModelIndexList &, const QPoint &))
    );
}

void ChangesetBrowser::receiveAll()
{
    changesetModel->receiveRevisions(true);
}

void ChangesetBrowser::receiveMore()
{
    changesetModel->receiveRevisions(false);
}

ChangesetBrowser::~ChangesetBrowser()
{
    branchesSplitter->saveState();
    changesetsSplitter->saveState();
    changeLogSplitter->saveState();

    delete changesetModel;
    delete revisionModel;
    if (branchModel) delete branchModel;
    Settings::setBool("ChangesetBrowserTree", tree);
}

void ChangesetBrowser::toggleTree()
{
    tree = !tree;
    displayBranchesAsTree->setText(
        !tree ? tr("display branches as tree") : tr("display branches flat")
    );
    initTreeWidget();
}

void ChangesetBrowser::branchesClicked(const QModelIndex & idx)
{
    QString branch = branchModel->data(idx, Qt::ToolTipRole).toString();
    changesetModel->setBranch(branch);
}

void ChangesetBrowser::changesetsClicked(const QModelIndex & idx)
{
    QModelIndex revIdx = changesetModel->index(idx.row(), 3, QModelIndex());
    QString revision = changesetModel->data(revIdx, Qt::DisplayRole).toString();
    changeLog->setHtml(changesetModel->data(revIdx, Qt::UserRole).toString());
    revisionModel->readDatabaseRevision(databaseFile, revision);
}

void ChangesetBrowser::initTreeWidget()
{
    if (branchModel)
    {
        disconnect(
            branchModel, SIGNAL(branchesRead(void)),
            this, SLOT(branchesRead(void))
        );
        delete branchModel;
    }

    branchModel = new Branches(this, databaseFile, tree);
    branchModel->readBranches();
    branches->setRootIsDecorated(false);
    branches->setModel(branchModel);

    connect(
        branchModel, SIGNAL(branchesRead(void)),
        this, SLOT(branchesRead(void))
    );
}

void ChangesetBrowser::branchesRead()
{
    branches->setExpanded(branchModel->index(0, 0, QModelIndex()), true);
}

void ChangesetBrowser::changesetsDoubleClicked(const QModelIndex & index)
{
    if (!index.isValid()) return;
    QModelIndex revIdx = changesetModel->index(index.row(), 3, QModelIndex());
    emit revisionManifest(revIdx.data().toString());
}

void ChangesetBrowser::contextMenuRequested(const QModelIndexList & indexList, const QPoint & pos)
{
    if (indexList.size() == 0) return;

    QModelIndex revIdx =
        changesetModel->index(indexList.at(0).row(), 3, QModelIndex());

    QString rev(revIdx.data().toString());

    QMenu menu(this);
    QFont activeFont;
    activeFont.setBold(true);

    QAction * actOpenManifest = menu.addAction(tr("View files in this revision"));
    QAction * actCopyRevisionId = menu.addAction(tr("Copy revision ID"));

    actOpenManifest->setFont(activeFont);

    QAction * act = menu.exec(pos);
    if (act == actOpenManifest)
    {
        emit revisionManifest(rev);
    }
    else
    if (act == actCopyRevisionId)
    {
        QClipboard * clipboard = QApplication::clipboard();
        clipboard->setText(rev);
    }
}

