/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "CommitRevision.h"
#include "Settings.h"
#include "MonotoneUtil.h"
#include "WorkspaceCommitter.h"

#include <QHeaderView>
#include <QMessageBox>
#include <QRegExp>
#include <QInputDialog>

CommitRevision::CommitRevision(QWidget * parent, const WorkspacePath & workspace)
  : Dialog(parent), workspacePath(workspace)
{
    setupUi(this);
    Dialog::init();

    revModel = new GetRevision(this);
    changeView->setModel(revModel);

    connect(
        revModel, SIGNAL(revisionRead()),
        this, SLOT(revisionRead())
    );

    connect(
        previousChangelogEntryList, SIGNAL(currentIndexChanged(int)),
        this, SLOT(setChangelogEntryFromList(int))
    );

    connect(
        changesAgainstParent, SIGNAL(currentIndexChanged(const QString &)),
        revModel, SLOT(showChangesAgainstParent(const QString &))
    );
}

CommitRevision::~CommitRevision()
{
    delete revModel;
}

void CommitRevision::readWorkspaceRevision()
{
    revModel->readWorkspaceRevision(workspacePath);

    QStringList entries = Settings::getItemList("ChangelogEntries");
    QRegExp re("\\s+");

    previousChangelogEntryList->clear();
    for (int i=0, j=entries.size(); i<j; i++)
    {
        QString entry = entries.at(i);
        QString shortened(entry);
        shortened.replace(re, " ");
        shortened = shortened.left(40);
        previousChangelogEntryList->insertItem(i, shortened, QVariant(entry));
    }
    previousChangelogEntryList->insertItem(-1, tr("-- select --"), QVariant());
}

void CommitRevision::setChangelogEntryFromList(int index)
{
    if (index < 0) return;
    changelogEntry->setPlainText(
        previousChangelogEntryList->itemData(index).toString()
    );
}

void CommitRevision::accept()
{
    Settings::addItemToList("ChangelogEntries", changelogEntry->toPlainText(), 10);

    QString branch = MonotoneUtil::getBranchName(workspacePath, QString());

    if (branch.isEmpty() && !checkAltBranch->isChecked())
    {
        QMessageBox::critical(
            this,
            tr("Workspace has no branch set"),
            tr("The workspace has no branch set. Please enter one."),
            QMessageBox::Ok
        );
        checkAltBranch->setChecked(true);
        return;
    }

    if (checkAltBranch->isChecked())
    {
        QString altBranchStr = altBranch->text().trimmed();
        if (altBranchStr.isEmpty())
        {
            QMessageBox::critical(
                this,
                tr("No branch specified"),
                tr("Please enter the name of an alternative branch."),
                QMessageBox::Ok
            );
            return;
        }
        branch = altBranchStr;
    }

    QString author;

    if (checkAltAuthor->isChecked())
    {
        QString altAuthorStr = altAuthor->text().trimmed();
        if (altAuthorStr.isEmpty())
        {
            QMessageBox::critical(
                this,
                tr("No author specified"),
                tr("Please enter the name of an alternative author."),
                QMessageBox::Ok
            );
            return;
        }
        author = altAuthorStr;
    }

    QStringList keys = MonotoneUtil::getPrivateKeyList(MonotoneUtil::getDatabaseFile(workspacePath));

    if (keys.size() == 0)
    {
        QMessageBox::critical(
            this,
            tr("No private key found"),
            tr("There are no private keys available with which to sign your "
               "commit. Please go to Database > Key Management and create a "
               "new private key first."),
            QMessageBox::Ok
        );
        return;
    }

    QString selectedKey = keys.at(0);

    if (keys.size() > 1)
    {
        bool ok;
        QString key = QInputDialog::getItem(
            this, tr("Please select a key"),
            tr("More than one private key is available.\n"
               "Please select the key with which you like to sign this commit:"),
            keys, 0, false, &ok);
        if (key.isEmpty() || !ok)
        {
            return;
        }
        selectedKey = key;
    }

    WorkspaceCommitter committer(workspacePath);

    if (!committer.run(selectedKey, branch, author, changelogEntry->toPlainText()))
    {
        QMessageBox::critical(
            this,
            tr("Could not commit revision"),
            tr("Unable to commit the revision - this may be a bug in guitone "
               "or a bug in monotone itself. Please take a closer look at the "
               "logs and optionally send in a bug report."),
            QMessageBox::Ok
        );
        done(0);
        return;
    }

    emit revisionCommitted(committer.getRevisionId());
    done(1);
}

void CommitRevision::revisionRead()
{
    if (revModel->rowCount(QModelIndex()) == 0)
    {
        QMessageBox::information(
            this,
            tr("No changes"),
            tr("The current workspace has no committable changes."),
            QMessageBox::Ok
        );
        reject();
    }

    changesAgainstParent->addItems(revModel->getParentRevisions());
}

