/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Preferences.h"
#include "MonotoneManager.h"
#include "Settings.h"
#include "DebugLog.h"
#include "Guitone.h"

#include <QFileDialog>
#include <QMessageBox>
#include <QHeaderView>

Preferences::Preferences(QWidget * parent) : Dialog(parent)
{
    setupUi(this);
    Dialog::init();

    mtnExecutablePath->setText(Settings::getMtnBinaryPath());

    enableConsoleLog->setCheckState(DebugLog::getConsoleLogEnabled() ?
                                    Qt::Checked : Qt::Unchecked);
    enableFileLog->setCheckState(DebugLog::getFileLogEnabled() ?
                                 Qt::Checked : Qt::Unchecked);

    enableFileLog->setText(enableFileLog->text().arg(DebugLog::logFilePath()));

    logLevel->addItem(tr("Very low (only fatal)"), 1);
    logLevel->addItem(tr("Low (critical)"), 2);
    logLevel->addItem(tr("Medium (warnings)"), 3);
    logLevel->addItem(tr("High (info messages)"), 4);

    int lvl = DebugLog::getLogLevel();
#ifndef QT_NO_DEBUG
    logLevel->addItem(tr("Very high (debug messages)"), 5);
#else
    lvl = lvl > 4 ? 4 : lvl;
#endif

    logLevel->setCurrentIndex(lvl - 1);

    checkForUpdates->setCheckState(
        Settings::getBool("CheckForUpdates", true) ?
        Qt::Checked : Qt::Unchecked
    );

    readWorkspaceIncrementally->setCheckState(
        Settings::getBool("ReadWorkspaceIncrementally", false) ?
        Qt::Checked : Qt::Unchecked
    );

    connect(
        selectMtnExecutable, SIGNAL(clicked()),
        this, SLOT(openFileBrowser())
    );

    relaxedVersionCheck->setCheckState(
        Settings::getBool("RelaxedVersionCheck", false) ?
        Qt::Checked : Qt::Unchecked
    );

    connect(
        relaxedVersionCheck, SIGNAL(toggled(bool)),
        this, SLOT(warnAboutRelaxedVersionChecking(bool))
    );
}

Preferences::~Preferences() {}

void Preferences::accept()
{
    QString oldPath = Settings::getMtnBinaryPath();
    QString newPath = mtnExecutablePath->text();
    bool checkUpperBound = relaxedVersionCheck->isChecked();

    QString installedVersion;

    // check if the new executable satisfies our version needs
    // we'll first check the program's plain version and if this
    // does not succeed we check for the interface version
    if (!APP->manager()->setMonotoneBinaryPath(newPath, checkUpperBound, installedVersion))
    {
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("The path to the monotone binary is either invalid "
               "or points to a version of monotone with which guitone can't "
               "work with. Guitone requires monotone with an interface version "
               "between '%1' and '%2'. Your installation has the interface "
               "version '%3'.")
                   .arg(MonotoneManager::MinInterfaceVersion)
                   .arg(MonotoneManager::MaxInterfaceVersion)
                   .arg(installedVersion),
            QMessageBox::Ok, 0, 0
        );
        return;
    }

    if (oldPath != newPath)
    {
        QMessageBox::information(
             this,
             tr("Notice"),
             tr("You need to restart guitone to use the new monotone binary."),
             QMessageBox::Ok
        );
        Settings::setMtnBinaryPath(newPath);
    }

    bool consoleEnabled = enableConsoleLog->isChecked();
    bool fileEnabled = enableFileLog->isChecked();
    int level = logLevel->itemData(logLevel->currentIndex()).toInt();

    DebugLog::setConsoleLogEnabled(consoleEnabled);
    DebugLog::setFileLogEnabled(fileEnabled);
    DebugLog::setLogLevel(level);

    Settings::setBool("CheckForUpdates", checkForUpdates->isChecked());
    Settings::setBool("RelaxedVersionCheck", relaxedVersionCheck->isChecked());

    if (Settings::getBool("ReadWorkspaceIncrementally", true) !=
        readWorkspaceIncrementally->isChecked())
    {
        QMessageBox::information(
             this,
             tr("Notice"),
             tr("You need to reload your workspace(s) in order to apply this new setting."),
             QMessageBox::Ok
        );
        Settings::setBool("ReadWorkspaceIncrementally",
                          readWorkspaceIncrementally->isChecked());
    }

    done(QDialog::Accepted);
}

void Preferences::openFileBrowser()
{
    QString file = QFileDialog::getOpenFileName(
         this,
         tr("Choose the monotone executable"),
         "",
         tr("Binaries (mtn mtn.exe)")
    );

    if (!file.isEmpty())
    {
        mtnExecutablePath->setText(file);
    }
}

void Preferences::warnAboutRelaxedVersionChecking(bool enabled)
{
    if (!enabled) return;

    QMessageBox::warning(
         this,
         tr("Warning"),
         tr("Activating this may let guitone work with newer, yet untested "
            "versions of monotone, in which certain functionalities may "
            "have changed or be no longer available.\n\n"
            "Use this at your own risk!"),
         QMessageBox::Ok
    );
}

