/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "UpdateWorkspace.h"
#include "SelectRevision.h"
#include "MonotoneUtil.h"
#include "Settings.h"

#include <QMessageBox>

UpdateWorkspace::UpdateWorkspace(QWidget * parent, const WorkspacePath & ws)
    : Dialog(parent), workspacePath(ws)
{
    setupUi(this);
    Dialog::init();

    updateToLatest->setText(
        updateToLatest->text().arg(MonotoneUtil::getBranchName(workspacePath))
    );

    DatabaseFile db = MonotoneUtil::getDatabaseFile(workspacePath);

    branchesModel = new Branches(this, db, false);
    branchesModel->readBranches();
    branchList->setModel(branchesModel);

    tagsModel = new Tags(this, db);
    tagsModel->readTags(QString());
    tagList->setModel(tagsModel);

    selectorCompleter = new QCompleter(
         Settings::getItemList("RecentUpdateSelectors"),
         this
    );

    selectorCompleter->setCaseSensitivity(Qt::CaseInsensitive);
    selectedRevision->setCompleter(selectorCompleter);

    connect(
        selectRevision, SIGNAL(clicked()),
        this, SLOT(slotSelectRevision())
    );

    connect(
        branchList, SIGNAL(activated(int)),
        this, SLOT(selectBranchOption())
    );

    connect(
        tagList, SIGNAL(activated(int)),
        this, SLOT(selectTagOption())
    );

    connect(
        selectedRevision, SIGNAL(textChanged(const QString &)),
        this, SLOT(selectSelectorOption())
    );
}

UpdateWorkspace::~UpdateWorkspace()
{
    delete branchesModel;
    delete tagsModel;
    delete selectorCompleter;
}

void UpdateWorkspace::accept()
{
    QString selector;

    if (updateToLatest->isChecked())
    {
        selector = "h:" + MonotoneUtil::getBranchName(workspacePath);
    }

    if (updateToBranch->isChecked())
    {
        QString selection(branchList->currentText());
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No branch selected"),
                tr("Please select a branch from the list of branches."),
                QMessageBox::Ok
             );
             return;
        }
        selector = "h:" + selection;
    }

    if (updateToTag->isChecked())
    {
        QString selection(tagList->currentText());
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No tag selected"),
                tr("Please select a tag from the list of tags (if any are available)."),
                QMessageBox::Ok
             );
             return;
        }
        selector = "t:" + selection;
    }

    if (updateToRevision->isChecked())
    {
        QString selection = selectedRevision->text();
        if (selection.isEmpty())
        {
            QMessageBox::warning(
                this,
                tr("No selector given"),
                tr("Please enter a monotone selector or choose a revision by clicking on 'select revision'."),
                QMessageBox::Ok
             );
             return;
        }
        selector = selection;
        Settings::addItemToList("RecentUpdateSelectors", selection, 10);
    }

    QStringList revList = MonotoneUtil::resolveSelector(
        MonotoneUtil::getDatabaseFile(workspacePath),
        selector
    );

    if (revList.size() == 0)
    {
        QMessageBox::warning(
            this,
            tr("No revisions matched"),
            tr("Your selection didn't match any revisions or was invalid."),
            QMessageBox::Ok
         );
         return;
    }

    if (revList.size() > 1)
    {
        QMessageBox::information(
            this,
            tr("Multiple revisions matched"),
            tr("Your selection matched multiple revisions. Please choose one."),
            QMessageBox::Ok
        );

        emit selectRevisionFromSelector(selector);
        return;
    }

    selectedRevisionId = revList.at(0);
    D(QString("Selected %1 as update revision").arg(selectedRevisionId));
    done(0);
}

void UpdateWorkspace::slotSelectRevision()
{
    emit selectRevisionFromSelector(QString());
}

void UpdateWorkspace::setSelectedRevision(const QString & revision)
{
    selectedRevision->setText(revision);
}

void UpdateWorkspace::selectBranchOption()
{
    updateToBranch->setChecked(true);
}

void UpdateWorkspace::selectTagOption()
{
    updateToTag->setChecked(true);
}

void UpdateWorkspace::selectSelectorOption()
{
    updateToRevision->setChecked(true);
}

